/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QAction>
#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QLabel>
#include <QListWidget>
#include <QListWidgetItem>
#include <QMenu>
#include <QMenuBar>
#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QWidget>
#include <QHBoxLayout>
#include <QSpacerItem>
#include <QFile>
#include <QFileDialog>
#include <QTextStream>
#include <QByteArray>
#include <QCloseEvent>
#include <QScreen>
#include <QSettings>
#include <QThread>
#include <QPainter>
#include <QPrinter>
#include <QMarginsF>
#include <QPageLayout>
#include <QToolBar>
#include <QDir>
#include <QPoint>
#include <QSize>
#include <QRect>
#include <QList>
#include <QFont>
#include <assert.h>

#include "../../core/messagebox.h"

#include "database.h"
#include "import_export.h"
#include "progressdialog.h"
#include "specificationeditordialog.h"
#include "specificationlimitsdialog.h"
#include "exportoptionsdialog.h"

#include "cmmspecifications.h"

TCMMSpecifications::TCMMSpecifications(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QAction								*about_action;
	QToolBar							*toolbar;
	QFrame								*separator_vline1;
	QFrame								*separator_hline1;
	QFrame								*separator_vline2;
	QGridLayout							*widget_layout;
	QHBoxLayout							*copyright_hlayout;
	QSpacerItem							*copyright_hspacer;
	QLabel								*make_label;
	QLabel								*model_label;
	QLabel								*option_label;
	QLabel								*specifications_label;
	QLabel								*logo_label;
	QLabel								*copyright_label;
	QMenu								*edit_menu;
	QMenu								*file_menu;
	QMenu								*help_menu;
	QMenuBar							*menu_bar;
	QTreeWidgetItem						*d_specification_tree_header_item;
	QWidget								*central_widget;
	QDir								data_dir;
	QPoint								position;
	QSize								size;
	QRect								desktop_rect;
	QRect								window_rect;
	QSizePolicy							sizepolicy_expanding_preferred(QSizePolicy::Expanding, QSizePolicy::Preferred);
	QFont								copyright_font;
	
	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	sizepolicy_expanding_preferred.setHorizontalStretch(0);
	sizepolicy_expanding_preferred.setVerticalStretch(0);
	
	d_application_path = qApp->applicationDirPath();
	
#ifdef Q_OS_MAC
	int index = d_application_path.lastIndexOf(QStringLiteral(".app/"));
	d_application_path.truncate(index);
	
	index = d_application_path.lastIndexOf('/');
	d_application_path.truncate(index);
#endif
	
#ifdef Q_OS_WIN
	d_data_path = d_application_path;
#else
	d_data_path = QDir::homePath();
#endif
	
	data_dir.setPath(d_data_path);
	
	if(!data_dir.exists(QStringLiteral(".cmmspecifications")))
	{
		data_dir.mkdir(QStringLiteral(".cmmspecifications"));
	}
	
	data_dir.cd(QStringLiteral(".cmmspecifications"));
	d_data_path = data_dir.absolutePath();

	d_settings = new QSettings(d_data_path + QStringLiteral("/cmmspecifications_settings.ini"),QSettings::IniFormat,this);

	position = d_settings->value("Mainwindow_Position", QPoint(30,30)).toPoint();
	size = d_settings->value("Mainwindow_Size", QSize(901,601)).toSize();
	
	desktop_rect = QGuiApplication::primaryScreen()->availableGeometry();
	
	window_rect.moveTopLeft(position);
	window_rect.setSize(size);
	
	window_rect = window_rect & desktop_rect;
	
	if(window_rect.isValid() && window_rect.width() > 500 && window_rect.height() > 500)
	{
		this->resize(window_rect.size());
		this->move(window_rect.topLeft());
	}
	else
	{
		this->resize(901,601);
		this->move(QPoint(30,30));
	}
	
	d_current_path = d_settings->value("Current_Path", QDir::homePath()).toString();
	
	data_dir.setPath(d_current_path);
	
	if(!data_dir.exists())
	{
#ifdef Q_OS_WIN
		d_current_path = d_application_path;
#else
		d_current_path = QDir::homePath();
#endif
	}

	d_new_database_action = new QAction(this);
	d_new_database_action->setIcon(QIcon(":/actions/new_database.png"));
	
	d_close_database_action = new QAction(this);
	d_close_database_action->setIcon(QIcon(":/actions/close_database.png"));

	d_open_database_action = new QAction(this);
	d_open_database_action->setIcon(QIcon(":/actions/open_database.png"));

	d_export_action = new QAction(this);
	d_export_action->setIcon(QIcon(":/actions/export.png"));
	
	d_import_action = new QAction(this);
	d_import_action->setIcon(QIcon(":/actions/import.png"));
	
	d_quit_action = new QAction(this);
	
	d_output_action = new QAction(this);
	d_output_action->setIcon(QIcon(":/actions/output.png"));
	
	d_delete_make_action = new QAction(this);
	d_delete_model_action = new QAction(this);
	d_delete_option_action = new QAction(this);

	d_edit_specifications_action = new QAction(this);
	d_edit_specifications_action->setIcon(QIcon(":/actions/edit.png"));
	
	d_limits_action = new QAction(this);
	d_limits_action->setIcon(QIcon(":/actions/limits.png"));

	about_action = new QAction(this);
	about_action->setIcon(QIcon(":/actions/about.png"));

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	widget_layout = new QGridLayout(central_widget);
	
	make_label = new QLabel(central_widget);
	widget_layout->addWidget(make_label,0,0,1,1);
	
	d_make_list = new QListWidget(central_widget);
	d_make_list->setSortingEnabled(true);
	d_make_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	sizepolicy_expanding_preferred.setHeightForWidth(d_make_list->sizePolicy().hasHeightForWidth());
	d_make_list->setSizePolicy(sizepolicy_expanding_preferred);
	widget_layout->addWidget(d_make_list,1,0,1,1);
	
	separator_vline1 = new QFrame(central_widget);
	separator_vline1->setFrameShape(QFrame::VLine);
	separator_vline1->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_vline1,0,1,2,1);

	model_label = new QLabel(central_widget);
	widget_layout->addWidget(model_label,0,2,1,1);
	
	d_model_list = new QListWidget(central_widget);
	d_model_list->setSortingEnabled(true);
	sizepolicy_expanding_preferred.setHeightForWidth(d_model_list->sizePolicy().hasHeightForWidth());
	d_model_list->setSizePolicy(sizepolicy_expanding_preferred);
	widget_layout->addWidget(d_model_list,1,2,1,1);
	
	separator_vline2 = new QFrame(central_widget);
	separator_vline2->setFrameShape(QFrame::VLine);
	separator_vline2->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_vline2,0,3,2,1);

	option_label = new QLabel(central_widget);
	widget_layout->addWidget(option_label,0,4,1,1);

	d_option_list = new QListWidget(central_widget);
	d_option_list->setSortingEnabled(true);
	sizepolicy_expanding_preferred.setHeightForWidth(d_option_list->sizePolicy().hasHeightForWidth());
	d_option_list->setSizePolicy(sizepolicy_expanding_preferred);
	widget_layout->addWidget(d_option_list,1,4,1,1);
	
	separator_hline1 = new QFrame(central_widget);
	separator_hline1->setFrameShape(QFrame::HLine);
	separator_hline1->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_hline1,2,0,1,5);

	specifications_label = new QLabel(central_widget);
	widget_layout->addWidget(specifications_label,3,0,1,5);

	d_specification_tree = new QTreeWidget(central_widget);

	d_specification_tree_header_item = d_specification_tree->headerItem();
	widget_layout->addWidget(d_specification_tree,4,0,1,5);
	
	copyright_hlayout = new QHBoxLayout();
	
	copyright_label = new QLabel(this);
	copyright_label->setFont(copyright_font);
	copyright_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	copyright_hlayout->addWidget(copyright_label);

	copyright_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	copyright_hlayout->addItem(copyright_hspacer);
	
	logo_label = new QLabel(this);
	logo_label->setMinimumSize(100,40);
	logo_label->setMaximumSize(100,40);
	copyright_hlayout->addWidget(logo_label);

	widget_layout->addLayout(copyright_hlayout,5,0,1,5);

	menu_bar = new QMenuBar(this);
	this->setMenuBar(menu_bar);

	file_menu = new QMenu(menu_bar);
	menu_bar->addMenu(file_menu);
	file_menu->addAction(d_new_database_action);
	file_menu->addAction(d_open_database_action);
	file_menu->addAction(d_close_database_action);
	file_menu->addSeparator();
	file_menu->addAction(d_export_action);
	file_menu->addAction(d_import_action);
	file_menu->addSeparator();
	file_menu->addAction(d_output_action);
	file_menu->addSeparator();
	file_menu->addAction(d_quit_action);

	edit_menu = new QMenu(menu_bar);
	menu_bar->addMenu(edit_menu);
	edit_menu->addAction(d_edit_specifications_action);
	edit_menu->addAction(d_limits_action);
	edit_menu->addSeparator();
	edit_menu->addAction(d_delete_make_action);
	edit_menu->addAction(d_delete_model_action);
	edit_menu->addAction(d_delete_option_action);
	edit_menu->addSeparator();

	help_menu = new QMenu(menu_bar);
	menu_bar->addMenu(help_menu);
	help_menu->addAction(about_action);
	menu_bar->resize(629,22);
	menu_bar->addAction(file_menu->menuAction());
	menu_bar->addAction(edit_menu->menuAction());
	menu_bar->addAction(help_menu->menuAction());
	
	toolbar = new QToolBar(this);
	toolbar->setIconSize(QSize(32,32));
	this->addToolBar(Qt::TopToolBarArea,toolbar);
	
	toolbar->addAction(d_new_database_action);
	toolbar->addAction(d_open_database_action);
	toolbar->addAction(d_close_database_action);
	toolbar->addSeparator();
	toolbar->addAction(d_export_action);
	toolbar->addAction(d_import_action);
	toolbar->addAction(d_output_action);
	toolbar->addSeparator();
	toolbar->addAction(d_edit_specifications_action);
	toolbar->addAction(d_limits_action);
	toolbar->addSeparator();
	toolbar->addAction(about_action);

	
	d_database = new TDatabase;
	d_msg_box = new TMessageBox(this);
	d_progress_dialog = new TProgressDialog(this);
	d_import_export_thread = new TImportExportThread(this);
	d_specification_editor_dialog = 0;
	d_specification_limits_dialog = 0;
	d_export_options_dialog = 0;

	// defaults
	d_delete_make_action->setEnabled(false);
	d_delete_model_action->setEnabled(false);
	d_delete_option_action->setEnabled(false);
	d_edit_specifications_action->setEnabled(false);
	d_limits_action->setEnabled(false);
	d_close_database_action->setEnabled(false);
	d_export_action->setEnabled(false);
	d_import_action->setEnabled(false);
	d_output_action->setEnabled(false);
	
	logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));

	this->setWindowTitle(QStringLiteral("CMM Specifications"));
	
	d_specification_tree_header_item->setText(0,QStringLiteral("Standard"));
	d_specification_tree_header_item->setText(1,QStringLiteral("Component"));
	d_specification_tree_header_item->setText(2,QStringLiteral("Expression"));
	d_specification_tree_header_item->setText(3,QStringLiteral("Source"));
	d_specification_tree_header_item->setText(4,QStringLiteral("Description"));
	
	d_new_database_action->setText(QStringLiteral("New Database..."));
	d_open_database_action->setText(QStringLiteral("Open Database..."));
	d_export_action->setText(QStringLiteral("Export..."));
	d_import_action->setText(QStringLiteral("Import..."));
	d_output_action->setText(QStringLiteral("Output PDF"));
	d_quit_action->setText(QStringLiteral("Quit"));
	d_close_database_action->setText(QStringLiteral("Close Database"));
	d_delete_make_action->setText(QStringLiteral("Delete Make"));
	d_delete_model_action->setText(QStringLiteral("Delete Model"));
	d_delete_option_action->setText(QStringLiteral("Delete Option"));
	d_edit_specifications_action->setText(QStringLiteral("Specification Editor..."));
	d_limits_action->setText(QStringLiteral("Limits..."));
	copyright_label->setText(QStringLiteral("Copyright (C) 2023 Select Calibration Incorporated.  www.selectcalibration.ca"));

	about_action->setText(QStringLiteral("About"));
	make_label->setText(QStringLiteral("Make:"));
	model_label->setText(QStringLiteral("Model:"));
	option_label->setText(QStringLiteral("Option:"));
	specifications_label->setText(QStringLiteral("Specifications:"));
	file_menu->setTitle(QStringLiteral("File"));
	edit_menu->setTitle(QStringLiteral("Edit"));
	help_menu->setTitle(QStringLiteral("Help"));

	connect(d_new_database_action,&QAction::triggered,this,&TCMMSpecifications::New_Database);
	connect(d_open_database_action,&QAction::triggered,this,&TCMMSpecifications::Open_Database);
	connect(d_close_database_action,&QAction::triggered,this,&TCMMSpecifications::Close_Database);
	connect(d_export_action,&QAction::triggered,this,&TCMMSpecifications::Export);
	connect(d_import_action,&QAction::triggered,this,&TCMMSpecifications::Import);
	connect(d_output_action,&QAction::triggered,this,&TCMMSpecifications::Output);
	connect(d_edit_specifications_action,&QAction::triggered,this,&TCMMSpecifications::Edit_Specification);
	connect(d_limits_action,&QAction::triggered,this,&TCMMSpecifications::Specification_Limits);
	connect(d_delete_make_action,&QAction::triggered,this,&TCMMSpecifications::Delete_Make);
	connect(d_delete_model_action,&QAction::triggered,this,&TCMMSpecifications::Delete_Model);
	connect(d_delete_option_action,&QAction::triggered,this,&TCMMSpecifications::Delete_Option);
	connect(about_action,&QAction::triggered,this,&TCMMSpecifications::Help_About);
	connect(d_quit_action,&QAction::triggered,this,&TCMMSpecifications::close);
	
	connect(d_import_export_thread,&QThread::started,this,&TCMMSpecifications::Thread_Started);
	connect(d_import_export_thread,&QThread::finished,this,&TCMMSpecifications::Thread_Finished);
	connect(d_import_export_thread,&TImportExportThread::Update_Progress,this,&TCMMSpecifications::Thread_Progress);
	connect(d_import_export_thread,&TImportExportThread::Update_Progress_Text,this,&TCMMSpecifications::Thread_Progress_Text);

	connect(d_make_list,&QListWidget::itemSelectionChanged,this,&TCMMSpecifications::Make_Selection_Changed);
	connect(d_model_list,&QListWidget::itemSelectionChanged,this,&TCMMSpecifications::Model_Selection_Changed);
	connect(d_option_list,&QListWidget::itemSelectionChanged,this,&TCMMSpecifications::Option_Selection_Changed);

}

TCMMSpecifications::~TCMMSpecifications(void)
{
	delete d_database;
	d_database = 0;
}

void TCMMSpecifications::New_Database(void)
{
	QString								file_name;
	QStringList							list;
	int									index;
	bool								valid;
	bool								existing_database;
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("New database"),
											 d_current_path,
											 QStringLiteral("Database File (*.db)"));
	
	if(file_name.length())
	{
		existing_database = QFile::exists(file_name);
		
		if(this->Load_Database(file_name))
		{
			if(existing_database)
			{
				d_msg_box->setText(QStringLiteral("Clear Database"));
				d_msg_box->setInformativeText(QStringLiteral("Do you want to clear the existing database?"));
				d_msg_box->setDetailedText(QString());
				d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
				d_msg_box->setDefaultButton(QMessageBox::No);
				d_msg_box->setIcon(QMessageBox::Question);
				
				if(QMessageBox::Yes == d_msg_box->exec())
				{
					d_database->Clear();
				}
			}
			
			d_edit_specifications_action->setEnabled(true);
			d_close_database_action->setEnabled(true);
			d_export_action->setEnabled(true);
			d_import_action->setEnabled(true);
		}
		
		d_current_path = file_name;
		
		index = d_current_path.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_current_path.truncate(index);
		}
		
		d_make_list->clear();
		d_model_list->clear();
		d_option_list->clear();
		d_specification_tree->clear();
		
		list = d_database->CMM_Make_List(&valid);
		
		if(valid)
		{
			d_make_list->addItems(list);
		}
		
		d_delete_make_action->setEnabled(false);
		d_delete_model_action->setEnabled(false);
		d_delete_option_action->setEnabled(false);
		d_limits_action->setEnabled(false);
	}
}

void TCMMSpecifications::Open_Database(void)
{
	QString								file_name;
	QStringList							list;
	int									index;
	bool								valid;
	
	file_name = QFileDialog::getOpenFileName(this,
										QStringLiteral("Open database"),
										d_current_path,
										QStringLiteral("Database File (*.db)"));

	if(file_name.length())
	{
		if(this->Load_Database(file_name))
		{
			d_edit_specifications_action->setEnabled(true);
			d_close_database_action->setEnabled(true);
			d_export_action->setEnabled(true);
			d_import_action->setEnabled(true);
		}
		
		d_current_path = file_name;
		
		index = d_current_path.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_current_path.truncate(index);
		}

		d_make_list->clear();
		d_model_list->clear();
		d_option_list->clear();
		d_specification_tree->clear();
		
		list = d_database->CMM_Make_List(&valid);
		
		if(valid)
		{
			d_make_list->addItems(list);
		}
		
		d_delete_make_action->setEnabled(false);
		d_delete_model_action->setEnabled(false);
		d_delete_option_action->setEnabled(false);
		d_limits_action->setEnabled(false);
	}
}

void TCMMSpecifications::Close_Database(void)
{
	d_database->Disconnect();
	
	d_make_list->clear();
	d_model_list->clear();
	d_option_list->clear();
	d_specification_tree->clear();
	
	d_edit_specifications_action->setEnabled(false);
	d_close_database_action->setEnabled(false);
	d_export_action->setEnabled(false);
	d_import_action->setEnabled(false);
	
	d_delete_make_action->setEnabled(false);
	d_delete_model_action->setEnabled(false);
	d_delete_option_action->setEnabled(false);
	d_limits_action->setEnabled(false);
	d_output_action->setEnabled(false);
}

void TCMMSpecifications::Export(void)
{
	QList<QListWidgetItem*>				items;
	QList<QListWidgetItem*>::const_iterator	item_iter;
	QString								file_name;
	QStringList							export_list;
	int									index;
	
	if(!d_database->Is_Connected())
	{
		d_msg_box->setText(QStringLiteral("Database Error"));
		d_msg_box->setInformativeText(QStringLiteral("Database not open."));
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return;
	}
	
	if(d_import_export_thread->isRunning())
	{
		d_msg_box->setText(QStringLiteral("IO Thread Running"));
		d_msg_box->setInformativeText(QStringLiteral("Database is currently being imported or exported."));
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return;
	}
	
	
	items = d_make_list->selectedItems();
	d_import_export_thread->Clear_Export_Make_List();
	
	if(items.size())
	{
		if(!d_export_options_dialog)
		{
			d_export_options_dialog = new TExportOptionsDialog(this);
		}
		
		if(QDialog::Accepted != d_export_options_dialog->exec())
		{
			return;
		}
		
		if(d_export_options_dialog->Export_Type() == TExportOptionsDialog::EXPORT_SELECTED_SPECIFICATIONS)
		{
			for(item_iter = items.begin();item_iter != items.end();++item_iter)
			{
				export_list.push_back((*item_iter)->text());
			}
			
			d_import_export_thread->Set_Export_Make_List(export_list);
		}
	}
		
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Export database"),
											 d_current_path,
											 QStringLiteral("Comma Separated File (*.csv)"));
	
	if(file_name.length())
	{
		d_current_path = file_name;

		index = d_current_path.lastIndexOf('/');

		if(!(index < 0))
		{
			d_current_path.truncate(index);
		}
		
		d_import_export_thread->Set_Database(d_database);
		d_import_export_thread->Set_Export_Mode();
		d_import_export_thread->Set_File_Name(file_name);
		
		d_import_export_thread->start(QThread::NormalPriority);
		
//		this->Export_Database(file_name);
	}
}

void TCMMSpecifications::Import(void)
{
	QString								file_name;
	int									index;
	
	if(!d_database->Is_Connected())
	{
		d_msg_box->setText(QStringLiteral("Database Error"));
		d_msg_box->setInformativeText(QStringLiteral("Database not open."));
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return;
	}
	
	if(d_import_export_thread->isRunning())
	{
		d_msg_box->setText(QStringLiteral("IO Thread Running"));
		d_msg_box->setInformativeText(QStringLiteral("Database is currently being imported or exported."));
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return;
	}

	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Import database"),
											 d_current_path,
											 QStringLiteral("Comma Separated File (*.csv)"));
	
	if(file_name.length())
	{
		d_current_path = file_name;
		
		index = d_current_path.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_current_path.truncate(index);
		}
		
		d_make_list->clear();
		d_model_list->clear();
		d_option_list->clear();
		d_specification_tree->clear();
		
		d_import_export_thread->Set_Database(d_database);
		d_import_export_thread->Set_Import_Mode();
		d_import_export_thread->Set_File_Name(file_name);
		
		d_import_export_thread->start(QThread::NormalPriority);
	}
}

void TCMMSpecifications::Output(void)
{
	QString								file_name;
	QPainter							painter;
	QPrinter							printer;//QPrinter::HighResolution);
	QRect								page_rect;
	QRect								rect;
	int									line_height;
	int									column_width;
	int									max_lines;
	QFont								text_font("Arial",9,QFont::Normal);
	QFont								title_font("Arial",14,QFont::Bold);
	QFont								machine_font("Arial",12,QFont::Bold);
	QFont								heading_font("Arial",10,QFont::Bold);
	QFont								footer_font("Arial",6,QFont::Normal);
	QPen								line_pen(QColor(128,128,128));
	QPen								text_pen(QColor(0,0,0));
	QBrush								title_background_brush(QColor(245,245,245));
	QMarginsF							margins;
	int									row,column;
	int									page_margin;
	bool								init_page;
	int									page_number;
	QString								name;
	QString								footer_text;
	int									column_positions[24];
	int									row_positions[9];
	QStringList							list;
	std::vector<QStringList>			specs;
	std::vector<QStringList>::iterator	iter;
	QString								make;
	QString								model;
	QString								option;
	QListWidgetItem						*item;
	TDatabaseItem						database_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	int									count;
	int									cntr;
	bool								valid;
	
	item = d_make_list->currentItem();
	
	if(!item)
	{
		return;
	}
	
	make = item->text();
	
	item = d_model_list->currentItem();
	
	if(!item)
	{
		return;
	}
	
	model = item->text();
	
	item = d_option_list->currentItem();
	
	if(!item)
	{
		return;
	}
	
	option = item->text();

	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Output specification"),
											 d_current_path,
											 QStringLiteral("Adobe PDF files (*.pdf)"));
	
	if(!file_name.length())
	{
		return;
	}
	
	if(!file_name.endsWith(QStringLiteral(".pdf"),Qt::CaseInsensitive))
	{
		file_name += QStringLiteral(".pdf");
	}
	
//	printer.setOrientation(QPrinter::Portrait);	// depreciated
	printer.setPageOrientation(QPageLayout::Portrait);
	printer.setOutputFormat(QPrinter::PdfFormat);
	printer.setOutputFileName(file_name);
	
//	printer.setPageMargins(0.0,0.0,0.0,0.0,QPrinter::Millimeter);	// depreciated

	margins.setLeft(0.0);
	margins.setTop(0.0);
	margins.setRight(0.0);
	margins.setBottom(0.0);
	printer.setPageMargins(margins,QPageLayout::Millimeter);
	
	//	printer.setColorMode(QPrinter::GrayScale);
	//	printer.setResolution(96);
	
	if(!painter.begin(&printer))
	{
		return;
	}
	
	footer_text = QStringLiteral("Specification Editor (C) Select Calibration Incorporated");
	
	painter.setFont(text_font);
//	page_rect = printer.pageRect();		// obsolete function
	page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
	
	page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.04);	// estimation based on 8.5x11 and reasonable margin
	page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
	
	rect = painter.boundingRect(0,0,0,0,Qt::AlignLeft,"10000.0000j");
	line_height = rect.height();
	
	// calculate column and row positions
	// component (3)  expression (6)  source(2)  description(remaining)
	//
	
	column_width = static_cast<int>((static_cast<double>(page_rect.width()) / 24.0) + 0.5);
	column_positions[0] = page_rect.left();
	
	for(column = 1;column < 24;column++)
	{
		column_positions[column] = column_positions[column - 1] + column_width;
	}
	
	// 0 title (3 rows)
	// 1 heading 1 (2 rows)
	// 2 heading 2 (2 rows)
	// 6 top of data
	// 7 footer
	
	row_positions[0] = page_rect.top();
	row_positions[1] = row_positions[0] + 3 * line_height;
	row_positions[2] = row_positions[1] + 1.5 * line_height;
	row_positions[3] = row_positions[2] + 1.5 * line_height;
	row_positions[4] = row_positions[3] + 1.5 * line_height;
	row_positions[5] = row_positions[4] + 1.5 * line_height;
	row_positions[6] = row_positions[5] + 1.5 * line_height;
	row_positions[7] = row_positions[6] + 1.5 * line_height;
	row_positions[8] = page_rect.bottom() - line_height;
	
	max_lines = ((row_positions[8] - row_positions[7]) / line_height) - 2;
	
	specs.clear();
	
	database_item = d_database->Database_Item(make,model,option,QStringLiteral("10360"),&valid);
	count = database_item.Specification_Items_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);
		
		if(!cntr)
		{
			specs.push_back(QStringList() << QStringLiteral("  ASME B89.4.10360, ISO/IEC 10360"));
		}
		
		list.clear();
		
		list.push_back(specification_item.component);
		list.push_back(specification_item.expression_item.Expression(TExpressionItem::UNITS_UM));
		
		if(specification_item.source == TDatabaseItem::SOURCE_MANUFACTURER)
		{
			list.push_back(QStringLiteral("MFG"));
		}
		else
		{
			list.push_back(QStringLiteral("EST"));
		}
		
		list.push_back(specification_item.description);
					   
		specs.push_back(list);
	}
	
	database_item = d_database->Database_Item(make,model,option,QStringLiteral("VDIVDE_2617"),&valid);
	count = database_item.Specification_Items_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);
		
		if(!cntr)
		{
			specs.push_back(QStringList() << QStringLiteral("    VDIVDE_2617"));
		}
		
		list.clear();
		
		list.push_back(specification_item.component);
		list.push_back(specification_item.expression_item.Expression(TExpressionItem::UNITS_UM));
		
		if(specification_item.source == TDatabaseItem::SOURCE_MANUFACTURER)
		{
			list.push_back(QStringLiteral("MFG"));
		}
		else
		{
			list.push_back(QStringLiteral("EST"));
		}
		
		list.push_back(specification_item.description);
		
		specs.push_back(list);
	}

	database_item = d_database->Database_Item(make,model,option,QStringLiteral("B89.4.1"),&valid);
	count = database_item.Specification_Items_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);
		
		if(!cntr)
		{
			specs.push_back(QStringList() << QStringLiteral("    ASME B89.4.1"));
		}
		
		list.clear();
		
		list.push_back(specification_item.component);
		list.push_back(specification_item.expression_item.Expression(TExpressionItem::UNITS_UM));
		
		if(specification_item.source == TDatabaseItem::SOURCE_MANUFACTURER)
		{
			list.push_back(QStringLiteral("MFG"));
		}
		else
		{
			list.push_back(QStringLiteral("EST"));
		}
		
		list.push_back(specification_item.description);
		
		specs.push_back(list);
	}

	database_item = d_database->Database_Item(make,model,option,QStringLiteral("CMMA"),&valid);
	count = database_item.Specification_Items_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);
		
		if(!cntr)
		{
			specs.push_back(QStringList() << QStringLiteral("    CMMA"));
		}
		
		list.clear();
		
		list.push_back(specification_item.component);
		list.push_back(specification_item.expression_item.Expression(TExpressionItem::UNITS_UM));
		
		if(specification_item.source == TDatabaseItem::SOURCE_MANUFACTURER)
		{
			list.push_back(QStringLiteral("MFG"));
		}
		else
		{
			list.push_back(QStringLiteral("EST"));
		}
		
		list.push_back(specification_item.description);
		
		specs.push_back(list);
	}
	
	page_number = 1;
	
	row = 1;
	init_page = true;
	
	for(iter = specs.begin();iter < specs.end();iter++)
	{
		if(init_page)
		{
			painter.setFont(title_font);
			painter.setPen(text_pen);
			
			rect.setRect(column_positions[0],row_positions[0],column_positions[23] - column_positions[0] + column_width,row_positions[1] - row_positions[0]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Coordinate Measuring Machine Specifications"));
			
			painter.setPen(line_pen);
			painter.drawLine(rect.bottomLeft(),rect.bottomRight());
			painter.setPen(text_pen);
			
			painter.setFont(machine_font);
			
			rect.setRect(column_positions[0],row_positions[2],3 * column_width,row_positions[3] - row_positions[2]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Make:"));
			
			rect.setRect(column_positions[3],row_positions[2],20 * column_width,row_positions[3] - row_positions[2]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,make);
			
			rect.setRect(column_positions[0],row_positions[3],3 * column_width,row_positions[4] - row_positions[3]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Model:"));
			
			rect.setRect(column_positions[3],row_positions[3],20 * column_width,row_positions[4] - row_positions[3]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,model);
			
			rect.setRect(column_positions[0],row_positions[4],3 * column_width,row_positions[5] - row_positions[4]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Option:"));
			
			rect.setRect(column_positions[3],row_positions[4],20 * column_width,row_positions[5] - row_positions[4]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,option);
			
			painter.setFont(machine_font);
			
			rect.setRect(column_positions[0],row_positions[6],4 * column_width, row_positions[7] - row_positions[6]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Component"));
			
			rect.setRect(column_positions[4],row_positions[6],11 * column_width, row_positions[7] - row_positions[6]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Expression"));
			
			rect.setRect(column_positions[12],row_positions[6],2 * column_width, row_positions[7] - row_positions[6]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Source"));
			
			rect.setRect(column_positions[15],row_positions[6],9 * column_width, row_positions[7] - row_positions[6]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,QString("Description"));
			
			painter.setPen(line_pen);
			
			rect.setRect(column_positions[0],row_positions[6],column_positions[23] - column_positions[0] + column_width,row_positions[7] - row_positions[6]);
			painter.drawLine(rect.bottomLeft(),rect.bottomRight());
			
			painter.setPen(text_pen);
			painter.setFont(footer_font);
			
			rect.setRect(column_positions[0],row_positions[8],20 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,footer_text);
			
			rect.setRect(column_positions[22],row_positions[8],2 * column_width, line_height);
			painter.drawText(rect,Qt::AlignRight,QString("Page %1").arg(page_number));
			
			rect.setRect(column_positions[0],row_positions[8] - line_height / 4,column_positions[23] - column_positions[0] + column_width,line_height);
			painter.drawLine(rect.topLeft(),rect.topRight());
			
			painter.setFont(text_font);
			init_page = false;
		}
		
		if((*iter).size() == 1)	// section title
		{
			painter.setFont(heading_font);
			
			rect.setRect(column_positions[0],row_positions[7] + row * line_height,24 * column_width,2 * line_height);
			painter.fillRect(rect,title_background_brush);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignVCenter,(*iter).at(0));
			
			row += 2;
			
			painter.setFont(text_font);
		}
		else if((*iter).size() == 4)
		{
			rect.setRect(column_positions[1],row_positions[7] + row * line_height,4 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,(*iter).at(0));
			
			rect.setRect(column_positions[4],row_positions[7] + row * line_height,11 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,(*iter).at(1));
			
			rect.setRect(column_positions[13],row_positions[7] + row * line_height,2 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,(*iter).at(2));
			
			rect.setRect(column_positions[15],row_positions[7] + row * line_height,9 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,(*iter).at(3));
			
			row++;
			
		}
		
		if(row > max_lines)
		{
			printer.newPage();
			page_number++;
			
			painter.setFont(text_font);
//			page_rect = printer.pageRect();		// depreciated
			page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
			
			page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.04);
			page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
			
			init_page = true;
			row = 1;
		}
	}
	
	
	row++;
	
	painter.setPen(line_pen);
	
	rect.setRect(column_positions[0],row_positions[7] + row * line_height,column_positions[23] - column_positions[0] + column_width,line_height);
	painter.drawLine(rect.bottomLeft(),rect.bottomRight());
	
	if((max_lines - row) > 8)	// only put legend on page if it fits
	{
		row++;
		
		painter.setPen(text_pen);
		painter.setFont(heading_font);
		
		rect.setRect(column_positions[0],row_positions[7] + row * line_height,4 * column_width,2 * line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::AlignBottom,"Legend");
		
		row += 2;
		painter.setFont(text_font);

		rect.setRect(column_positions[1],row_positions[7] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"L");
		rect.setRect(column_positions[4],row_positions[7] + row * line_height,21 * column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"Length in mm");
		row++;
		
		rect.setRect(column_positions[1],row_positions[7] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"Td");
		rect.setRect(column_positions[4],row_positions[7] + row * line_height,21 * column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"Temperature deviation from standard reference in Celsius");
		row++;
		
		rect.setRect(column_positions[1],row_positions[7] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"MFG");
		rect.setRect(column_positions[4],row_positions[7] + row * line_height,21 * column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"Specification provided by manufacturer");
		row++;
		
		rect.setRect(column_positions[1],row_positions[7] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"EST");
		rect.setRect(column_positions[4],row_positions[7] + row * line_height,21 * column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft,"Estimated specification derived from manufacturer specifications");
		row++;
		row++;
		
		painter.setFont(text_font);
		
		rect.setRect(column_positions[0],row_positions[7] + row * line_height,23 * column_width,2 * line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::AlignBottom,"All expression results in um (micrometers) unless otherwise specified.");
	}
	
	painter.end();
}

void TCMMSpecifications::Delete_Make(void)
{
	QString								make;
	QStringList							list;
	QList<QListWidgetItem*>				items;
	QList<QListWidgetItem*>::const_iterator item_iter;
	bool								prev_state;
	bool								valid;
	bool								delete_error;
	
	items = d_make_list->selectedItems();
	
	if(items.size())
	{
		d_msg_box->setText(QStringLiteral("Delete Current Make"));
		d_msg_box->setInformativeText(QStringLiteral("Are you sure you want to delete selected Make and all sub items?"));
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			return;
		}
		
		delete_error = false;
		
		for(item_iter = items.begin();item_iter != items.end();++item_iter)
		{
			make = (*item_iter)->text();
			
			if(!d_database->Delete_Make(make))
			{
				delete_error = true;
			}
		}
		
		if(delete_error)
		{
			d_msg_box->setText(QStringLiteral("Delete Error"));
			d_msg_box->setInformativeText(QStringLiteral("Cannot delete selected Make."));
			d_msg_box->setDetailedText(QString());
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			d_msg_box->setIcon(QMessageBox::Warning);
			
			d_msg_box->exec();
		}
		
		
		// update with new entries
		list = d_database->CMM_Make_List(&valid);
		prev_state = d_make_list->blockSignals(true);
		d_make_list->clear();
		d_make_list->addItems(list);
		d_make_list->blockSignals(prev_state);
		
		prev_state = d_model_list->blockSignals(true);
		d_model_list->clear();
		d_model_list->blockSignals(prev_state);
		
		prev_state = d_option_list->blockSignals(true);
		d_option_list->clear();
		d_option_list->blockSignals(prev_state);

		d_specification_tree->clear();
		
		d_delete_make_action->setEnabled(false);
		d_delete_model_action->setEnabled(false);
		d_delete_option_action->setEnabled(false);
		d_limits_action->setEnabled(false);
		d_output_action->setEnabled(false);
	}
}

void TCMMSpecifications::Delete_Model(void)
{
	QString								make;
	QString								model;
	QStringList							list;
	QList<QListWidgetItem*>				items;
	bool								prev_state;
	bool								valid;

	items = d_make_list->selectedItems();
	
	if(items.size())
	{
		make = items[0]->text();
		
		items = d_model_list->selectedItems();
		
		if(items.size())
		{
			model = items[0]->text();
			
			d_msg_box->setText(QStringLiteral("Delete Current Model"));
			d_msg_box->setInformativeText(QStringLiteral("Are you sure you want to delete selected Model and all sub items?"));
			d_msg_box->setDetailedText(QString());
			d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			d_msg_box->setDefaultButton(QMessageBox::No);
			d_msg_box->setIcon(QMessageBox::Question);
			
			if(QMessageBox::No == d_msg_box->exec())
			{
				return;
			}
			
			if(!d_database->Delete_Model(make,model))
			{
				d_msg_box->setText(QStringLiteral("Delete Error"));
				d_msg_box->setInformativeText(QStringLiteral("Cannot delete selected Model."));
				d_msg_box->setDetailedText(d_database->Last_Error());
				d_msg_box->setStandardButtons(QMessageBox::Ok);
				d_msg_box->setDefaultButton(QMessageBox::Ok);
				d_msg_box->setIcon(QMessageBox::Warning);
				
				d_msg_box->exec();
				
			}
			
			
			list = d_database->CMM_Model_List(make,&valid);
			prev_state = d_model_list->blockSignals(true);
			d_model_list->clear();
			d_model_list->addItems(list);
			d_model_list->blockSignals(prev_state);
			
			prev_state = d_option_list->blockSignals(true);
			d_option_list->clear();
			d_option_list->blockSignals(prev_state);

			d_specification_tree->clear();
			
			d_delete_model_action->setEnabled(false);
			d_delete_option_action->setEnabled(false);
			d_limits_action->setEnabled(false);
			d_output_action->setEnabled(false);
		}
	}
}

void TCMMSpecifications::Delete_Option(void)
{
	QString								make;
	QString								model;
	QString								option;
	QStringList							list;
	QList<QListWidgetItem*>				items;
	bool								prev_state;
	bool								valid;

	items = d_make_list->selectedItems();
	
	if(items.size())
	{
		make = items[0]->text();
		
		items = d_model_list->selectedItems();
		
		if(items.size())
		{
			model = items[0]->text();
			
			items = d_option_list->selectedItems();
			
			if(items.size())
			{
				option = items[0]->text();

				d_msg_box->setText(QStringLiteral("Delete Current Option"));
				d_msg_box->setInformativeText(QStringLiteral("Are you sure you want to delete selected Option and all sub items?"));
				d_msg_box->setDetailedText(QString());
				d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
				d_msg_box->setDefaultButton(QMessageBox::No);
				d_msg_box->setIcon(QMessageBox::Question);
				
				if(QMessageBox::No == d_msg_box->exec())
				{
					return;
				}
				
				if(!d_database->Delete_Option(make,model,option))
				{
					d_msg_box->setText(QStringLiteral("Delete Error"));
					d_msg_box->setInformativeText(QStringLiteral("Cannot delete selected Option."));
					d_msg_box->setDetailedText(d_database->Last_Error());
					d_msg_box->setStandardButtons(QMessageBox::Ok);
					d_msg_box->setDefaultButton(QMessageBox::Ok);
					d_msg_box->setIcon(QMessageBox::Warning);
					
					d_msg_box->exec();

				}
				
				list = d_database->CMM_Option_List(make,model,&valid);
				prev_state = d_option_list->blockSignals(true);
				d_option_list->clear();
				d_option_list->addItems(list);
				d_option_list->blockSignals(prev_state);
				
				d_specification_tree->clear();
				
				d_delete_option_action->setEnabled(false);
				d_limits_action->setEnabled(false);
				d_output_action->setEnabled(false);
			}
		}
	}
}

void TCMMSpecifications::Edit_Specification(void)
{
	QString								make;
	QString								model;
	QString								option;
	QString								description;
	QStringList							standard_type_list;
	QStringList							component_type_list;
	QStringList							list;
	QStringList::const_iterator			standard_type_iter;
	QStringList::const_iterator			component_type_iter;
	QListWidgetItem						*list_item;
	TSpecificationEditorDialog::TEntryItem entry_item;
	std::vector<TSpecificationEditorDialog::TEntryItem> entry_items;
	std::vector<TSpecificationEditorDialog::TEntryItem>::const_iterator entry_item_iter;
	TDatabaseItem						database_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	bool								prev_state;
	bool								valid;
	
	if(!d_specification_editor_dialog)
	{
		d_specification_editor_dialog = new TSpecificationEditorDialog(this);
	}
	
	list_item = d_make_list->currentItem();
	
	if(list_item)
	{
		if(list_item->isSelected())
		{
			make = list_item->text();
		}
	}
	
	list_item = d_model_list->currentItem();
	
	if(list_item)
	{
		if(list_item->isSelected())
		{
			model = list_item->text();
		}
	}
	
	list_item = d_option_list->currentItem();
	
	if(list_item)
	{
		if(list_item->isSelected())
		{
			option = list_item->text();
		}
	}
	
	standard_type_list = d_database->Standard_Type_List(&valid);
	
	if(valid)
	{
		for(standard_type_iter = standard_type_list.begin();standard_type_iter != standard_type_list.end();++standard_type_iter)
		{
			if(make.length() && model.length() && option.length())
			{
				database_item = d_database->Database_Item(make,model,option,(*standard_type_iter),&valid);
			}
			
			component_type_list = d_database->Component_Type_List((*standard_type_iter),&valid);
			entry_item.standard = (*standard_type_iter);

			if(valid)
			{
				for(component_type_iter = component_type_list.begin();component_type_iter != component_type_list.end();++component_type_iter)
				{
					entry_item.component = (*component_type_iter);
					entry_item.expression = database_item.Specification_Expression((*component_type_iter),TExpressionItem::UNITS_UM);
					entry_item.description = d_database->Component_Description((*standard_type_iter),(*component_type_iter));
					entry_item.expression_types = d_database->Component_Expression_Types((*standard_type_iter),(*component_type_iter));
					
					if(database_item.Specification_Source(*component_type_iter) == TDatabaseItem::SOURCE_ESTIMATED)
					{
						entry_item.expression = QString();
					}
					
					entry_items.push_back(entry_item);
				}
			}
		}
		
		d_specification_editor_dialog->Reset(d_database,make,model,option,entry_items);
		
		if(QDialog::Accepted == d_specification_editor_dialog->exec())
		{
			make = d_specification_editor_dialog->Make();
			model = d_specification_editor_dialog->Model();
			option = d_specification_editor_dialog->Option();
			
			database_item.Set_Make(make);
			database_item.Set_Model(model);
			database_item.Set_Option(option);
			
			entry_items = d_specification_editor_dialog->Specification_Entries();
			
			for(standard_type_iter = standard_type_list.begin();standard_type_iter != standard_type_list.end();++standard_type_iter)
			{
				database_item.Clear_Specification_Items();
				
				database_item.Set_Standard_Type(*standard_type_iter);
				
				for(entry_item_iter = entry_items.begin();entry_item_iter != entry_items.end();++entry_item_iter)
				{
					if((*entry_item_iter).standard.compare(*standard_type_iter) == 0)
					{
						specification_item.component = (*entry_item_iter).component;
						specification_item.expression_item.Set_Expression((*entry_item_iter).expression,TExpressionItem::UNITS_UM);
						specification_item.source = TDatabaseItem::SOURCE_MANUFACTURER;
												
						database_item.Add_Specification_Item(specification_item);
					}
				}
				
				d_database->Add_Database_Item(database_item);
			}
			
			d_database->Create_Estimated_Specifications(make,model,option);
			
			// update with new entries
			list = d_database->CMM_Make_List(&valid);
			prev_state = d_make_list->blockSignals(true);
			d_make_list->clear();
			d_make_list->addItems(list);
			d_make_list->blockSignals(prev_state);

			list = d_database->CMM_Model_List(make,&valid);
			prev_state = d_model_list->blockSignals(true);
			d_model_list->clear();
			d_model_list->addItems(list);
			d_model_list->blockSignals(prev_state);

			list = d_database->CMM_Option_List(make,model,&valid);
			prev_state = d_option_list->blockSignals(true);
			d_option_list->clear();
			d_option_list->addItems(list);
			d_option_list->blockSignals(prev_state);
			
			this->Set_Selection(make,model,option);
			
			d_delete_make_action->setEnabled(true);
			d_delete_model_action->setEnabled(true);
			d_delete_option_action->setEnabled(true);
			d_limits_action->setEnabled(true);
			d_output_action->setEnabled(true);
		}
	}
}

void TCMMSpecifications::Specification_Limits(void)
{
	QString								make;
	QString								model;
	QString								option;
	QStringList							standard_type_list;
	QStringList::const_iterator			standard_type_iter;
	QListWidgetItem						*list_item;
	TDatabaseItem						database_item;
	bool								valid;
	
	if(!d_specification_limits_dialog)
	{
		d_specification_limits_dialog = new TSpecificationLimitsDialog(this);
	}
	
	list_item = d_make_list->currentItem();
	
	if(list_item)
	{
		if(list_item->isSelected())
		{
			make = list_item->text();
		}
	}
	
	list_item = d_model_list->currentItem();
	
	if(list_item)
	{
		if(list_item->isSelected())
		{
			model = list_item->text();
		}
	}
	
	list_item = d_option_list->currentItem();
	
	if(list_item)
	{
		if(list_item->isSelected())
		{
			option = list_item->text();
		}
	}
	
	d_specification_limits_dialog->Reset(make,model,option);

	standard_type_list = d_database->Standard_Type_List(&valid);
	
	if(valid)
	{
		for(standard_type_iter = standard_type_list.begin();standard_type_iter != standard_type_list.end();++standard_type_iter)
		{
			if(make.length() && model.length() && option.length())
			{
				database_item = d_database->Database_Item(make,model,option,(*standard_type_iter),&valid);
				
				if(valid)
				{
					d_specification_limits_dialog->Add_Database_Item(database_item);
				}
			}
		}
		
		d_specification_limits_dialog->Update();
		d_specification_limits_dialog->exec();
	}
}

void TCMMSpecifications::Help_About(void)
{
	d_msg_box->setText(QStringLiteral("Specification Editor"));
	d_msg_box->setInformativeText(QStringLiteral("Specification Editor - Version 1.0"));
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	d_msg_box->setIcon(QMessageBox::Information);
	
	d_msg_box->exec();
}

void TCMMSpecifications::Make_Selection_Changed(void)
{
	QString								current_make_selection;
	QString								current_model_selection;
	QStringList							list;
	QList<QListWidgetItem*>				items;
	QListWidgetItem						*item;
	bool								valid;
	
	d_delete_model_action->setEnabled(false);
	d_delete_option_action->setEnabled(false);
	d_limits_action->setEnabled(false);
	d_output_action->setEnabled(false);

	items = d_make_list->selectedItems();
	
	d_delete_make_action->setEnabled(items.size());

	if(items.size() == 1)
	{
		current_make_selection = items[0]->text();

		item = d_model_list->currentItem();
		
		if(item)
		{
			current_model_selection = item->text();
		}
		
		d_model_list->clear();
		
		list = d_database->CMM_Model_List(current_make_selection,&valid);
		
		if(valid)
		{
			d_model_list->addItems(list);
			
			if(current_model_selection.length())
			{
				items = d_model_list->findItems(current_model_selection,Qt::MatchExactly);
				
				if(items.size())
				{
					d_model_list->setCurrentItem(items[0]);
				}
			}
		}
		
	}
	else
	{
		d_model_list->clear();
		d_option_list->clear();
		d_specification_tree->clear();
	}
}

void TCMMSpecifications::Model_Selection_Changed(void)
{
	QString								current_make_selection;
	QString								current_model_selection;
	QString								current_option_selection;
	QStringList							list;
	QList<QListWidgetItem*>				items;
	QListWidgetItem						*item;
	bool								valid;
	
	d_delete_model_action->setEnabled(false);
	d_delete_option_action->setEnabled(false);
	d_limits_action->setEnabled(false);
	d_output_action->setEnabled(false);

	items = d_make_list->selectedItems();
	
	if(items.size() != 1)
	{
		return;
	}
	
	current_make_selection = items[0]->text();
	
	items = d_model_list->selectedItems();
	
	if(items.size() == 1)
	{
		current_model_selection = items[0]->text();
		d_delete_model_action->setEnabled(true);
	
		item = d_option_list->currentItem();
		
		if(item)
		{
			current_option_selection = item->text();
		}
		
		d_option_list->clear();
		
		list = d_database->CMM_Option_List(current_make_selection,current_model_selection,&valid);
		
		if(valid)
		{
			d_option_list->addItems(list);
			
			if(current_option_selection.length())
			{
				items = d_option_list->findItems(current_option_selection,Qt::MatchExactly);
				
				if(items.size())
				{
					d_option_list->setCurrentItem(items[0]);
				}
			}
		}
	}
	else
	{
		d_option_list->clear();
		d_specification_tree->clear();
	}
}

void TCMMSpecifications::Option_Selection_Changed(void)
{
	TDatabaseItem						database_item;
	QString								current_make_selection;
	QString								current_model_selection;
	QString								current_option_selection;
	QStringList							list;
	QStringList::const_iterator			iter;
	QList<QListWidgetItem*>				items;
	bool								valid;
	
	d_delete_option_action->setEnabled(false);
	d_limits_action->setEnabled(false);
	d_output_action->setEnabled(false);

	items = d_make_list->selectedItems();
	
	if(!items.size())
	{
		return;
	}
	
	current_make_selection = items[0]->text();
	
	items = d_model_list->selectedItems();
	
	if(items.size() != 1)
	{
		return;
	}
	
	current_model_selection = items[0]->text();
	
	items = d_option_list->selectedItems();
	
	if(items.size())
	{
		current_option_selection = items[0]->text();
		d_delete_option_action->setEnabled(true);
		d_limits_action->setEnabled(true);
		d_output_action->setEnabled(true);

		d_specification_tree->clear();
		
		list = d_database->Standard_Type_List(&valid);
		
		if(valid)
		{
			for(iter = list.begin();iter != list.end();++iter)
			{
				database_item = d_database->Database_Item(current_make_selection,current_model_selection,current_option_selection,(*iter),&valid);
				
				if(valid)
				{
					this->Create_Specification_Entry((*iter),database_item);
				}
			}
		}
		
		d_specification_tree->expandAll();
		
		d_specification_tree->resizeColumnToContents(0);
		d_specification_tree->resizeColumnToContents(1);
		d_specification_tree->resizeColumnToContents(2);
		d_specification_tree->resizeColumnToContents(3);
		d_specification_tree->resizeColumnToContents(4);
	}
	else
	{
		d_specification_tree->clear();
	}
}

void TCMMSpecifications::Thread_Started(void)
{
	d_progress_dialog->Set_Range(100);
	d_progress_dialog->Set_Value(0);
		
	d_progress_dialog->show();
}

void TCMMSpecifications::Thread_Finished(void)
{
	QStringList							list;
	bool								valid;
	
	d_progress_dialog->hide();

	if(d_import_export_thread->Error_State())
	{
		d_msg_box->setText(QStringLiteral("Import Export Error"));
		d_msg_box->setInformativeText(QStringLiteral("Operation failed."));
		d_msg_box->setDetailedText(d_import_export_thread->Last_Error());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return;
	}
	
	if(d_import_export_thread->Execution_Mode() == TImportExportThread::MODE_IMPORT)
	{
		list = d_database->CMM_Make_List(&valid);
		
		if(valid)
		{
			d_make_list->addItems(list);
		}
	}
}

void TCMMSpecifications::Thread_Progress(
	int									value,
	int									range)
{
	d_progress_dialog->Set_Range(range);
	d_progress_dialog->Set_Value(value);
}

void TCMMSpecifications::Thread_Progress_Text(
	const QString						&text)
{
	d_progress_dialog->Set_Text(text);
}

void TCMMSpecifications::closeEvent(
	QCloseEvent 						*event)
{
	d_settings->setValue("Mainwindow_Position", this->pos());
	d_settings->setValue("Mainwindow_Size", this->size());
	d_settings->setValue("Current_Path", d_current_path);

	if(d_import_export_thread->isRunning())
	{
		event->ignore();
		return;
	}

	event->accept();
}


bool TCMMSpecifications::Load_Database(
	const QString						&database_file)
{
	QDir								dir;
	QString								path;
	
	if(d_import_export_thread->isRunning())
	{
		d_msg_box->setText(QStringLiteral("IO Thread Running"));
		d_msg_box->setInformativeText(QStringLiteral("Database is currently being imported or exported."));
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
		
		return false;
	}
	
	if(d_database->Is_Connected())
	{
		d_database->Disconnect();
	}
	
	if(d_database->Connect(database_file))
	{
		if(d_database->Initialize())
		{
			return true;
		}
		else
		{
			d_msg_box->setText(QStringLiteral("Initialize Error"));
			d_msg_box->setInformativeText(QStringLiteral("Error trying to initialize the database."));
			d_msg_box->setDetailedText(d_database->Last_Error());
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			d_msg_box->setIcon(QMessageBox::Warning);
			
			d_msg_box->exec();
		}
	}
	else
	{
		d_msg_box->setText(QStringLiteral("Connect Error"));
		d_msg_box->setInformativeText(QStringLiteral("Error trying to connect to the database."));
		d_msg_box->setDetailedText(d_database->Last_Error());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Warning);
		
		d_msg_box->exec();
	}
	
	return false;
}

void TCMMSpecifications::Set_Selection(
	const QString 						&make,
	const QString 						&model,
	const QString 						&option)
{
	QList<QListWidgetItem*>				items;
	TDatabaseItem						database_item;
	QStringList							list;
	QStringList::const_iterator			iter;
	bool								prev_state;
	bool								valid;
	
	items = d_make_list->findItems(make,Qt::MatchExactly);
	
	if(items.size())
	{
		prev_state = d_make_list->blockSignals(true);
		d_make_list->setCurrentItem(items[0],QItemSelectionModel::Select);
		d_make_list->blockSignals(prev_state);
	}
	else
	{
		return;
	}
	
	items = d_model_list->findItems(model,Qt::MatchExactly);
	
	if(items.size())
	{
		prev_state = d_model_list->blockSignals(true);
		d_model_list->setCurrentItem(items[0],QItemSelectionModel::Select);
		d_model_list->blockSignals(prev_state);
	}
	else
	{
		return;
	}
	
	items = d_option_list->findItems(option,Qt::MatchExactly);
	
	if(items.size())
	{
		prev_state = d_option_list->blockSignals(true);
		d_option_list->setCurrentItem(items[0],QItemSelectionModel::Select);
		d_option_list->blockSignals(prev_state);
	}
	else
	{
		return;
	}

	d_specification_tree->clear();
	
	list = d_database->Standard_Type_List(&valid);
	
	if(valid)
	{
		for(iter = list.begin();iter != list.end();++iter)
		{
			database_item = d_database->Database_Item(make,model,option,(*iter),&valid);
			
			if(valid)
			{
				this->Create_Specification_Entry((*iter),database_item);
			}
		}
	}
	
	d_specification_tree->expandAll();
	
	d_specification_tree->resizeColumnToContents(0);
	d_specification_tree->resizeColumnToContents(1);
	d_specification_tree->resizeColumnToContents(2);
	d_specification_tree->resizeColumnToContents(3);
	d_specification_tree->resizeColumnToContents(4);
	
}

void TCMMSpecifications::Create_Specification_Entry(
	const QString 						&standard,
	const TDatabaseItem 				&database_item)
{
	QTreeWidgetItem						*item;
	QTreeWidgetItem						*sub_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	int									count;
	int									cntr;
	
	item = new QTreeWidgetItem(d_specification_tree);
	
	item->setText(0,standard);
	
	count = database_item.Specification_Items_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);
		
		sub_item = new QTreeWidgetItem();
		
		sub_item->setText(1,specification_item.component);
		sub_item->setText(2,specification_item.expression_item.Expression(TExpressionItem::UNITS_UM));
		
		switch(specification_item.source)
		{
			case TDatabaseItem::SOURCE_MANUFACTURER:
				sub_item->setText(3,QStringLiteral("MFG"));
				break;
				
			case TDatabaseItem::SOURCE_ESTIMATED:
				sub_item->setText(3,QStringLiteral("EST"));
				break;
		}
		
		sub_item->setText(4,specification_item.description);
		
		item->addChild(sub_item);
	}
}


