/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QVariant>
#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QtSql/QSqlField>
#include <QtSql/QSqlDriver>
#include <QDate>
#include <set>
#include <assert.h>

#include <QFile>
#include <QDataStream>

#include "database.h"

TDatabase::TDatabase(void)
{
	d_database_instance_name = QStringLiteral("cmmspecifications_database");
	
	d_database = 0;
}

TDatabase::~TDatabase(void)
{
	this->Disconnect();
}

bool TDatabase::Is_Connected(void) const
{
	if(!d_database)
	{
		return false;
	}
	
	return d_database->isOpen();
}

bool TDatabase::CMM_Make_Exists(
	const QString						&make) const
{
	int									make_id;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}
	
	return true;
}

bool TDatabase::CMM_Model_Exists(
	const QString						&make,
	const QString						&model) const
{
	int									make_id;
	int									model_type_id;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}
	
	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return false;
	}

	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT component_id FROM specification WHERE make_id=%1 AND model_id=%2").arg(make_id).arg(model_type_id)))
	{
		if(query.next())
		{
			return true;
		}
	}
	
	return false;
}

bool TDatabase::CMM_Option_Exists(
	const QString						&make,
	const QString						&model,
	const QString						&option) const
{
	int									make_id;
	int									model_type_id;
	int									option_type_id;

	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}
	
	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return false;
	}
	
	option_type_id = this->Option_Type_Id(option);
	
	if(option_type_id < 0)
	{
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT component_id FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3")
				  .arg(make_id)
				  .arg(model_type_id)
				  .arg(option_type_id)))
	{
		if(query.next())
		{
			return true;
		}
	}

	
	return true;
}

bool TDatabase::Standard_Type_Exists(
	const QString						&standard_type) const
{
	int									standard_type_id;
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return false;
	}
	
	return true;
}

bool TDatabase::Component_Type_Exists(
	const QString						&standard_type,
	const QString						&component_type) const
{
	int									standard_type_id;
	int									component_type_id;
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return false;
	}
	
	component_type_id = this->Component_Type_Id(standard_type_id,component_type);
	
	if(component_type_id < 0)
	{
		return false;
	}
	
	return true;
}


QStringList TDatabase::CMM_Make_List(
	bool								* const valid) const
{
	QStringList							list;
	QString								text;
	
	assert(valid);
	*valid = false;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return list;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QStringLiteral("SELECT name FROM make")))
	{
		*valid = true;
		
		while(query.next())
		{
			text = query.value(0).toString();
			list.push_back(text);
		}
	}
	
	return list;
}

QStringList TDatabase::CMM_Model_List(
	const QString						&make,
	bool								* const valid) const
{
	QStringList							list;
	QString								text;
	std::set<int>						existing_model_types;
	std::set<int>::const_iterator		iter;
	int									make_id;
	int									model_type_id;
	
	assert(valid);
	*valid = false;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return list;
	}
	
	QSqlQuery							query(*d_database);

	if(query.exec(QString("SELECT model_id FROM specification WHERE make_id=%1").arg(make_id)))
	{
		while(query.next())
		{
			model_type_id = query.value(0).toInt();
			
			iter = existing_model_types.find(model_type_id);
			
			if((model_type_id < 0) == false && iter == existing_model_types.end())
			{
				*valid = true;
				existing_model_types.insert(existing_model_types.end(),model_type_id);
				list.push_back(this->Model_Type(model_type_id));
			}
		}
	}
	
	return list;
}

QStringList TDatabase::CMM_Option_List(
	const QString						&make,
	const QString						&model,
	bool								* const valid) const
{
	QStringList							list;
	QString								text;
	std::set<int>						existing_option_types;
	std::set<int>::const_iterator		iter;
	int									make_id;
	int									model_type_id;
	int									option_type_id;
	
	assert(valid);
	*valid = false;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return list;
	}
	
	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return list;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT option_id FROM specification WHERE make_id=%1 AND model_id=%2").arg(make_id).arg(model_type_id)))
	{
		while(query.next())
		{
			option_type_id = query.value(0).toInt();
			
			iter = existing_option_types.find(option_type_id);
			
			if((option_type_id < 0) == false && iter == existing_option_types.end())
			{
				*valid = true;
				existing_option_types.insert(existing_option_types.end(),option_type_id);
				list.push_back(this->Option_Type(option_type_id));
			}
		}
	}
	
	return list;
}

QStringList TDatabase::Standard_Type_List(
	bool								* const valid) const
{
	QStringList							list;
	
	assert(valid);
	*valid = false;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return list;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QStringLiteral("SELECT name FROM standard_types")))
	{
		*valid = true;
		
		while(query.next())
		{
			list.push_back(query.value(0).toString());
		}
	}
	
	return list;
}

QStringList TDatabase::Component_Type_List(
	const QString						&standard_type,
	bool								* const valid) const
{
	QStringList							list;
	int									standard_type_id;
	
	assert(valid);
	*valid = false;
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return list;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT name FROM component_types WHERE standard_id=%1").arg(standard_type_id)))
	{
		*valid = true;
		
		while(query.next())
		{
			list.push_back(query.value(0).toString());
		}
	}
	
	return list;
}

QString TDatabase::Component_Description(
	const QString 						&standard_type,
	const QString 						&component_name) const
{
	int									standard_type_id;
	int									component_type_id;
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return QString();
	}
	
	component_type_id = this->Component_Type_Id(standard_type_id,component_name);
	
	if(component_type_id < 0)
	{
		return QString();
	}

	return Component_Description(standard_type_id,component_type_id);
}

int TDatabase::Component_Expression_Types(
	const QString 						&standard_type,
	const QString 						&component_name) const
{
	int									standard_type_id;
	int									component_type_id;
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return TExpressionItem::EXPRESSION_UNKNOWN;
	}
	
	component_type_id = this->Component_Type_Id(standard_type_id,component_name);
	
	if(component_type_id < 0)
	{
		return TExpressionItem::EXPRESSION_UNKNOWN;
	}
	
	return Component_Expression_Types(standard_type_id,component_type_id);
}

TDatabaseItem TDatabase::Database_Item(
	const QString						&make,
	const QString						&model,
	const QString						&option,
	const QString						&standard_type,
	bool								* const valid) const
{
	TDatabaseItem						database_item;
	int									make_id;
	int									model_type_id;
	int									option_type_id;
	int									standard_type_id;
	int									component_type_id;
	bool								init;
	TDatabaseItem::TSpecification_Item	specification_item;
	
	assert(valid);
	*valid = false;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return database_item;
	}
	
	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return database_item;
	}
	
	option_type_id = this->Option_Type_Id(option);
	
	if(option_type_id < 0)
	{
		return database_item;
	}
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return database_item;
	}

	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT component_id,expression,source FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3 AND standard_id=%4")
				  .arg(make_id)
				  .arg(model_type_id)
				  .arg(option_type_id)
				  .arg(standard_type_id)))
	{
		*valid = true;
		init = true;
		
		while(query.next())
		{
			if(init)
			{
				database_item.Clear_Specification_Items();

				database_item.Set_Make(make);
				database_item.Set_Model(model);
				database_item.Set_Option(option);
				database_item.Set_Standard_Type(standard_type);
				
				init = false;
			}
			
			component_type_id = query.value(0).toInt();
			
			specification_item.component = this->Component_Name(standard_type_id,component_type_id);
			specification_item.expression_item.Set_Expression(query.value(1).toString(),TExpressionItem::UNITS_MM);
			specification_item.source = static_cast<TDatabaseItem::TSpecification_Item_Source>(query.value(2).toInt());
			specification_item.description = Component_Description(standard_type_id,component_type_id);
			specification_item.expression_types = Component_Expression_Types(standard_type_id,component_type_id);

			database_item.Add_Specification_Item(specification_item);
		}
	}
	else
	{
		d_last_error = query.lastError().text();
		return database_item;
	}

	return database_item;
}

bool TDatabase::Connect(
	const QString						&file_name)
{
	QStringList							list;
	QString								text;
	QList<QString>::const_iterator		iter;
	

	if(!QSqlDatabase::isDriverAvailable(QStringLiteral("QSQLITE")))
	{
		d_last_error = QStringLiteral("Database driver 'QSQLITE' not available");
		return false;
	}
	
	if(QSqlDatabase::contains(d_database_instance_name))
	{
		d_database = new QSqlDatabase(QSqlDatabase::database(d_database_instance_name));
	}
	else
	{
		d_database = new QSqlDatabase(QSqlDatabase::addDatabase(QStringLiteral("QSQLITE"),d_database_instance_name));
	}
	
	d_database->setDatabaseName(file_name);
	
	if(!d_database->open())
	{
		text = QString("Cannot open database.\n\tDatabase Error: '%1'\n\tDriver Error: '%2'\n")
									.arg(d_database->lastError().databaseText())
									.arg(d_database->lastError().driverText());
		
		text += QString("\t\tDatabase File Name: '%1'\n").arg(d_database_file_name);
		
		list = QSqlDatabase::drivers();
		
		text += QString("\tAvailable database drivers:\n");
		for(iter = list.begin();iter != list.end();++iter)
		{
			text += QString("\t\t%1\n").arg(*iter);
		}
		
		QSqlDatabase::removeDatabase(d_database_instance_name);
		
		delete d_database;
		d_database = 0;

		d_last_error = text;
		return false;
	}
	
	d_database_file_name = file_name;

	return true;
}

bool TDatabase::Initialize(void)
{
	QStringList							list;
	QString								text;
	QList<QString>::const_iterator		iter;

	if(!d_database)
	{
		d_last_error = QStringLiteral("Database not open");
		return false;
	}
	
	if(!d_database->isOpen())
	{
		d_last_error = QStringLiteral("Database not open");
		return false;
	}
	
	// create necessary tables if necessary
	QSqlQuery							query(*d_database);
	
	list = d_database->tables();
	
	if(!list.contains("make",Qt::CaseInsensitive))
	{
		text = QString("CREATE TABLE make(make_id INTEGER PRIMARY KEY AUTOINCREMENT,name TEXT)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'make(...)'");
			return false;
		}
	}
	
	if(!list.contains("model_types",Qt::CaseInsensitive))
	{
		text = QString("CREATE TABLE model_types(model_id INTEGER PRIMARY KEY AUTOINCREMENT,name TEXT)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'model_types(...)'");
			return false;
		}
	}
	
	if(!list.contains("option_types",Qt::CaseInsensitive))
	{
		text = QString("CREATE TABLE option_types(option_id INTEGER PRIMARY KEY AUTOINCREMENT,name TEXT)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'option_types(...)'");
			return false;
		}
	}
	
	if(!list.contains("standard_types",Qt::CaseInsensitive))
	{
		text = QString("CREATE TABLE standard_types(standard_type_id INTEGER PRIMARY KEY AUTOINCREMENT,name TEXT)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'standard_types(...)'");
			return false;
		}
	}
	
	if(!this->Initialize_Default_Standard_Types())
	{
		return false;
	}
	
	if(!list.contains("component_types",Qt::CaseInsensitive))
	{
		text = QString("CREATE TABLE component_types(component_type_id INTEGER PRIMARY KEY AUTOINCREMENT,standard_id INTEGER,name TEXT,description TEXT,expression_types INTEGER)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'component_types(...)'");
			return false;
		}
	}
	
	if(!this->Initialize_Default_Component_Types())
	{
		return false;
	}

	if(!list.contains("specification",Qt::CaseInsensitive))
	{
		text = QString("CREATE TABLE specification(make_id INTEGER,model_id INTEGER,option_id INTEGER,standard_id INTEGER,component_id INTEGER,expression TEXT,source INTEGER)");
		
		if(!query.exec(text))
		{
			d_last_error = QString("Cannot create table 'specification(...)'");
			return false;
		}		
	}

	return true;
}

void TDatabase::Disconnect(void)
{
	if(d_database)
	{
		d_database->close();
		
		delete d_database;
		d_database = 0;

		if(QSqlDatabase::contains(d_database_instance_name))
		{
			QSqlDatabase::removeDatabase(d_database_instance_name);
		}
	}
}

void TDatabase::Clear(void)
{
	if(d_database)
	{
		QSqlQuery							query(*d_database);
		
		query.exec(QStringLiteral("DELETE FROM make"));
		query.exec(QStringLiteral("DELETE FROM model_types"));
		query.exec(QStringLiteral("DELETE FROM option_types"));
		query.exec(QStringLiteral("DELETE FROM standard_types"));
		query.exec(QStringLiteral("DELETE FROM component_types"));
		query.exec(QStringLiteral("DELETE FROM specification"));
		
		this->Initialize_Default_Standard_Types();
		this->Initialize_Default_Component_Types();
	}
}

bool TDatabase::Add_Database_Item(
	const TDatabaseItem					&database_item)
{
	QString								command;
	int									make_id;
	int									model_type_id;
	int									option_type_id;
	int									standard_type_id;
	int									component_type_id;
	int									count;
	int									cntr;
	TDatabaseItem::TSpecification_Item	specification_item;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}

	QSqlQuery							query(*d_database);
	
	if(!this->CMM_Make_Exists(database_item.Make()))
	{
		command = QString("INSERT INTO make (name) VALUES ('%1')").arg(database_item.Make());
		
		if(!query.exec(command))
		{
			d_last_error = QString("Cannot add data to table 'make'.  Reason: %1").arg(query.lastError().text());
			return false;
		}
	}
	
	make_id = this->Make_Id(database_item.Make());

	
	if(!this->Model_Type_Exists(database_item.Model()))
	{
		command = QString("INSERT INTO model_types (name) VALUES ('%1')").arg(database_item.Model());
		
		if(!query.exec(command))
		{
			d_last_error = QString("Cannot add data to table 'model_types'.  Reason: %1").arg(query.lastError().text());
			return false;
		}
	}
	
	model_type_id = this->Model_Type_Id(database_item.Model());

	
	if(!this->Option_Type_Exists(database_item.Option()))
	{
		command = QString("INSERT INTO option_types (name) VALUES ('%1')").arg(database_item.Option());
		
		if(!query.exec(command))
		{
			d_last_error = QString("Cannot add data to table 'option_types'.  Reason: %1").arg(query.lastError().text());
			return false;
		}
	}
	
	option_type_id = this->Option_Type_Id(database_item.Option());
	
	standard_type_id = this->Standard_Type_Id(database_item.Standard_Type());
	
	if(standard_type_id < 0)
	{
		return false;
	}
	
	if(!query.exec(QString("DELETE FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3 AND standard_id=%4")
			   .arg(make_id)
			   .arg(model_type_id)
			   .arg(option_type_id)
			   .arg(standard_type_id)))
	{
		d_last_error = QString("Cannot delete existing specification data.  Reason: %1").arg(query.lastError().text());
		return false;
	}

	count = database_item.Specification_Items_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);

		component_type_id = this->Component_Type_Id(standard_type_id,specification_item.component);
		
		if(component_type_id < 0)
		{
			d_last_error = QString("Component type '%1' does not exist").arg(specification_item.component);
			return false;
		}
		
		if(!query.exec(QString("INSERT INTO specification (make_id,model_id,option_id,standard_id,component_id,expression,source) VALUES(%1,%2,%3,%4,%5,'%6',%7)")
					   .arg(make_id)
					   .arg(model_type_id)
					   .arg(option_type_id)
					   .arg(standard_type_id)
					   .arg(component_type_id)
					   .arg(specification_item.expression_item.Expression(TExpressionItem::UNITS_MM))
					   .arg(specification_item.source)))
		{
			d_last_error = QString("Cannot add data to table 'specification'.  Reason: %1").arg(query.lastError().text());
			return false;
		}
	}
	
	return true;
}

bool TDatabase::Create_Estimated_Specifications(
	const QString						&make,
	const QString						&model,
	const QString						&option)
{
	QStringList							list;
	QStringList::const_iterator			list_iter;
	int									make_id;
	int									model_type_id;
	int									option_type_id;
	TDatabaseItem						database_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	bool								valid;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}

	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return false;
	}

	option_type_id = this->Option_Type_Id(option);
	
	if(option_type_id < 0)
	{
		return false;
	}

	this->Remove_Estimated_Specifications(make_id,model_type_id,option_type_id);

	list = this->Standard_Type_List(&valid);
	
	if(!valid)
	{
		return false;
	}
	
	for(list_iter = list.begin();list_iter != list.end();++list_iter)
	{
		this->Create_Estimated_Specifications(make_id,model_type_id,option_type_id,*list_iter);
	}
	
	return true;
}

bool TDatabase::Delete_Make(
	const QString						&make)
{
	int									make_id;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	if(!query.exec(QString("DELETE FROM make WHERE make_id=%1").arg(make_id)))
	{
	   d_last_error = QString("Cannot delete make entry for '%1'").arg(make);
	   return false;
	}
	
	if(!query.exec(QString("DELETE FROM specification WHERE make_id=%1").arg(make_id)))
	{
		d_last_error = QString("Cannot delete specification entry for '%1'").arg(make);
		return false;
	}
	
	return true;
}

bool TDatabase::Delete_Model(
	const QString						&make,
	const QString						&model)
{
	int									make_id;
	int									model_type_id;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}
	
	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return false;
	}

	QSqlQuery							query(*d_database);
	
	if(!query.exec(QString("DELETE FROM specification WHERE make_id=%1 AND model_id=%2").arg(make_id).arg(model_type_id)))
	{
		d_last_error = QString("Cannot delete specification entry for '%1' and '%2'").arg(make).arg(model);
		return false;
	}
	
	return true;
}

bool TDatabase::Delete_Option(
	const QString						&make,
	const QString						&model,
	const QString						&option)
{
	int									make_id;
	int									model_type_id;
	int									option_type_id;
	
	make_id = this->Make_Id(make);
	
	if(make_id < 0)
	{
		return false;
	}
	
	model_type_id = this->Model_Type_Id(model);
	
	if(model_type_id < 0)
	{
		return false;
	}
	
	option_type_id = this->Option_Type_Id(option);
	
	if(option_type_id < 0)
	{
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	if(!query.exec(QString("DELETE FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3").arg(make_id).arg(model_type_id).arg(option_type_id)))
	{
		d_last_error = QString("Cannot delete specification entry for '%1' and '%2' and '%3'").arg(make).arg(model).arg(option);
		return false;
	}
	
	return true;
}

bool TDatabase::Initialize_Default_Standard_Types(void)
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}

	if(!this->Standard_Type_Exists(QStringLiteral("10360")))
	{
		if(!this->Add_Standard_Type(QStringLiteral("10360")))
		{
			return false;
		}
	}
	
	if(!this->Standard_Type_Exists(QStringLiteral("B89.4.1")))
	{
		if(!this->Add_Standard_Type(QStringLiteral("B89.4.1")))
		{
			return false;
		}
	}

	if(!this->Standard_Type_Exists(QStringLiteral("VDIVDE_2617")))
	{
		if(!this->Add_Standard_Type(QStringLiteral("VDIVDE_2617")))
		{
			return false;
		}
	}
	
	if(!this->Standard_Type_Exists(QStringLiteral("CMMA")))
	{
		if(!this->Add_Standard_Type(QStringLiteral("CMMA")))
		{
			return false;
		}
	}

	return true;
}

bool TDatabase::Add_Standard_Type(
	const QString						&standard)
{
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	if(!query.exec(QString("INSERT INTO standard_types (name) VALUES ('%1')").arg(standard)))
	{
		d_last_error = QString("Cannot add data to table 'standard_types'.  Reason: %1").arg(query.lastError().text());
		return false;
	}
	
	return true;
}

bool TDatabase::Initialize_Default_Component_Types(void)
{
	int									standard_type_id;
	int									component_type_id;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}
	
	standard_type_id = this->Standard_Type_Id(QStringLiteral("10360"));
	
	if(!(standard_type_id < 0))
	{
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Pftu"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Pftu"),
										 QStringLiteral("Maximum probe form error"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("R0"));
		
		if(component_type_id < 0)
		{
			
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("R0"),
										 QStringLiteral("Maximum length variation for E0"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Rpt"));
		
		if(component_type_id < 0)
		{
			
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Rpt"),
										 QStringLiteral("Maximum sphere position variation"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}

		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("E0"));
		
		if(component_type_id < 0)
		{
		
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("E0"),
										 QStringLiteral("Maximum length error"),
										 TExpressionItem::EXPRESSION_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT | TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("E150"));
		
		if(component_type_id < 0)
		{
			
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("E150"),
										 QStringLiteral("Maximum length error with probe offset"),
										 TExpressionItem::EXPRESSION_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT | TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Ts_Max"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Ts_Max"),
										 QStringLiteral("Maximum volume temperature celsius"),
										 TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Ts_Min"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Ts_Min"),
										 QStringLiteral("Minimum volume temperature celsius"),
										 TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS))
			{
				return false;
			}
		}

	}
	
	standard_type_id = this->Standard_Type_Id(QStringLiteral("B89.4.1"));
	
	if(!(standard_type_id < 0))
	{
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Repeat"));
		
		if(component_type_id < 0)
		{
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Repeat"),
										 QStringLiteral("Maximum sphere position variation"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Volume"));
		
		if(component_type_id < 0)
		{
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Volume"),
										 QStringLiteral("Maximum length variation of a ballbar"),
										 TExpressionItem::EXPRESSION_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Probe_Roll"));
		
		if(component_type_id < 0)
		{
			
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Probe_Roll"),
										 QStringLiteral("Maximum third axis ballbar length variation"),
										 TExpressionItem::EXPRESSION_VALUE_LENGTH| TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH))
			{
				return false;
			}
		}
	}

	standard_type_id = this->Standard_Type_Id(QStringLiteral("VDIVDE_2617"));
	
	if(!(standard_type_id < 0))
	{
	
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("U1"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("U1"),
										 QStringLiteral("Maximum length error in one axis"),
										 TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("U2"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("U2"),
										 QStringLiteral("Maximum length error in two axis"),
										 TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("U3"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("U3"),
										 QStringLiteral("Maximum length error in three axis"),
										 TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("V1"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("V1"),
										 QStringLiteral("Maximum probe error in one axis"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("V2"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("V2"),
										 QStringLiteral("Maximum probe error in two axis"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("V3"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("V3"),
										 QStringLiteral("Maximum probe error in three axis"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
	}
	
	standard_type_id = this->Standard_Type_Id(QStringLiteral("CMMA"));
	
	if(!(standard_type_id < 0))
	{
	
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("G"));
		
		if(component_type_id < 0)
		{
			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("G"),
										 QStringLiteral("Maximum probe error"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("M"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("M"),
										 QStringLiteral("Maximum length measurement error"),
										 TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT | TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE))
			{
				return false;
			}
		}
		
		component_type_id = this->Component_Type_Id(standard_type_id,QStringLiteral("Repeat"));
		
		if(component_type_id < 0)
		{

			if(!this->Add_Component_Type(standard_type_id,
										 QStringLiteral("Repeat"),
										 QStringLiteral("Maximum length variation error"),
										 TExpressionItem::EXPRESSION_VALUE))
			{
				return false;
			}
		}
	}
	
	return true;
}

bool TDatabase::Add_Component_Type(
	const int							standard_type_id,
	const QString						&component,
	const QString 						&description,
	const int							expression_types)
{
	QString								command;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	command = QString("INSERT INTO component_types (standard_id,name,description,expression_types) VALUES (%1,'%2','%3',%4)")
						.arg(standard_type_id)
						.arg(component)
						.arg(description)
						.arg(expression_types);
	
	if(!query.exec(command))
	{
		d_last_error = QString("Cannot add data to table 'component_types'.  Reason: %1").arg(query.lastError().text());		
		return false;
	}
	
	return true;
}

bool TDatabase::Model_Type_Exists(
	const QString						&model) const
{
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT model_id FROM model_types WHERE name='%1'").arg(model)))
	{
		if(query.next())
		{
			return true;
		}
	}
	
	return false;
}

bool TDatabase::Option_Type_Exists(
	const QString						&option) const
{
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT option_id FROM option_types WHERE name='%1'").arg(option)))
	{
		if(query.next())
		{
			return true;
		}
	}
	
	return false;
}

int TDatabase::Make_Id(
	const QString						&make) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return -1;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT make_id FROM make WHERE name='%1'").arg(make)))
	{
		if(query.next())
		{
			return query.value(0).toInt();
		}
	}
	
	d_last_error = QStringLiteral("CMM Make not found.");
	return -1;
}

int TDatabase::Model_Type_Id(
	const QString						&model) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return -1;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT model_id FROM model_types WHERE name='%1'").arg(model)))
	{
		if(query.next())
		{
			return query.value(0).toInt();
		}
	}
	
	d_last_error = QStringLiteral("CMM Model not found.");
	return -1;
}

int TDatabase::Option_Type_Id(
	const QString						&option) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return -1;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT option_id FROM option_types WHERE name='%1'").arg(option)))
	{
		if(query.next())
		{
			return query.value(0).toInt();
		}
	}
	
	d_last_error = QStringLiteral("CMM Option not found.");
	return -1;
}

int TDatabase::Standard_Type_Id(
	const QString						&standard_type) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return -1;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT standard_type_id FROM standard_types WHERE name='%1'").arg(standard_type)))
	{
		if(query.next())
		{
			return query.value(0).toInt();
		}
	}
	
	d_last_error = QStringLiteral("Standard Type not found.");
	return -1;
}


QString TDatabase::Make(
	const int							make_id) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return QString();
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT name FROM make WHERE make_id=%1").arg(make_id)))
	{
		if(query.next())
		{
			return query.value(0).toString();
		}
	}
	
	d_last_error = QStringLiteral("CMM Make not found.");
	return QString();
}

QString TDatabase::Model_Type(
	const int							model_type_id) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return QString();
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT name FROM model_types WHERE model_id=%1").arg(model_type_id)))
	{
		if(query.next())
		{
			return query.value(0).toString();
		}
	}

	d_last_error = QStringLiteral("CMM Model not found.");
	return QString();
}

QString TDatabase::Option_Type(
		const int						option_type_id) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return QString();
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT name FROM option_types WHERE option_id=%1").arg(option_type_id)))
	{
		if(query.next())
		{
			return query.value(0).toString();
		}
	}
	
	d_last_error = QStringLiteral("CMM Option not found.");
	return QString();
}

QString TDatabase::Standard_Type(
	const int							standard_type_id) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return QString();
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT name FROM standard_types where standard_type_id=%1").arg(standard_type_id)))
	{
		if(query.next())
		{
			return query.value(0).toString();
		}
	}
	
	d_last_error = QStringLiteral("Standard Type not found.");
	return QString();
}

int TDatabase::Component_Type_Id(
	const int							standard_type_id,
	const QString						&component) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return -1;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT component_type_id FROM component_types WHERE standard_id=%1 AND name='%2'").arg(standard_type_id).arg(component)))
	{
		if(query.next())
		{
			return query.value(0).toInt();
		}
	}
	
	d_last_error = QStringLiteral("Component Type not found.");
	return -1;
}

QString TDatabase::Component_Name(
	const int 							standard_type_id,
	const int 							component_type_id) const
{	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return QString();
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT name FROM component_types WHERE standard_id=%1 AND component_type_id=%2").arg(standard_type_id).arg(component_type_id)))
	{
		if(query.next())
		{
			return query.value(0).toString();
		}
	}
	
	d_last_error = QStringLiteral("Component Type not found.");
	return QString();
}

QString TDatabase::Component_Description(
	const int 							standard_type_id,
	const int 							component_type_id) const
{	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return QString();
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT description FROM component_types WHERE standard_id=%1 AND component_type_id=%2").arg(standard_type_id).arg(component_type_id)))
	{
		if(query.next())
		{
			return query.value(0).toString();
		}
	}
	
	d_last_error = QStringLiteral("Component Type not found.");
	return QString();
}

int TDatabase::Component_Expression_Types(
	const int 							standard_type_id,
	const int 							component_type_id) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return TExpressionItem::EXPRESSION_UNKNOWN;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT expression_types FROM component_types WHERE standard_id=%1 AND component_type_id=%2").arg(standard_type_id).arg(component_type_id)))
	{
		if(query.next())
		{
			return query.value(0).toInt();
		}
	}
	
	d_last_error = QStringLiteral("Component Type not found.");
	return TExpressionItem::EXPRESSION_UNKNOWN;
}

bool TDatabase::Specification_Exists(
	const int							make_id,
	const int							model_id,
	const int							option_type_id,
	const int							standard_type_id,
	const int							component_id) const
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT source FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3 AND standard_id=%4 AND component_id=%5")
				  .arg(make_id)
				  .arg(model_id)
				  .arg(option_type_id)
				  .arg(standard_type_id)
				  .arg(component_id)))
	{
		if(query.next())
		{
			return true;
		}
	}
	
	return false;
}

TDatabaseItem TDatabase::Database_Item(
	const int							make_id,
	const int							model_id,
	const int							option_type_id,
	const int							standard_type_id,
	bool								* const valid) const
{
	TDatabaseItem						database_item;
	int									component_id;
	bool								init;
	TDatabaseItem::TSpecification_Item	specification_item;
	
	assert(valid);
	
	*valid = false;
	
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return database_item;
	}
	
	QSqlQuery							query(*d_database);
	
	if(query.exec(QString("SELECT component_id,expression,source FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3 AND standard_id=%4")
				  .arg(make_id)
				  .arg(model_id)
				  .arg(option_type_id)
				  .arg(standard_type_id)))
	{		
		*valid = true;
		
		init = true;
		
		while(query.next())
		{			
			if(init)
			{
				database_item.Clear_Specification_Items();
				
				database_item.Set_Make(this->Make(make_id));
				database_item.Set_Model(this->Model_Type(model_id));
				database_item.Set_Option(this->Option_Type(option_type_id));
				database_item.Set_Standard_Type(this->Standard_Type(standard_type_id));
				
				init = false;
			}
			
			component_id = query.value(0).toInt();
			
			specification_item.component = this->Component_Name(standard_type_id,component_id);
			specification_item.expression_item.Set_Expression(query.value(1).toString(),TExpressionItem::UNITS_MM);
			specification_item.source = static_cast<TDatabaseItem::TSpecification_Item_Source>(query.value(2).toInt());
			
			database_item.Add_Specification_Item(specification_item);
		}
	}
	else
	{
		d_last_error = query.lastError().text();
		return database_item;
	}
	
	return database_item;
}

bool TDatabase::Remove_Estimated_Specifications(
	const int							make_id,
	const int							model_id,
	const int							option_type_id)
{
	if(!this->Is_Connected())
	{
		d_last_error = QStringLiteral("Not connected");
		return false;
	}
	
	QSqlQuery							query(*d_database);
	
	if(!query.exec(QString("DELETE FROM specification WHERE make_id=%1 AND model_id=%2 AND option_id=%3 AND source=%4")
			   .arg(make_id)
			   .arg(model_id)
			   .arg(option_type_id)
			   .arg(TDatabaseItem::SOURCE_ESTIMATED)))
	{
		d_last_error = QStringLiteral("Cannot delete specification entry");
		return false;
	}

	return true;

}

bool TDatabase::Create_Estimated_Specifications(
	const int							make_id,
	const int							model_id,
	const int							option_type_id,
	const QString						&standard_type)
{
	TDatabaseItem						database_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	std::vector<TDatabaseItem::TSpecification_Item> new_b89_items;
	std::vector<TDatabaseItem::TSpecification_Item> new_10360_items;
	std::vector<TDatabaseItem::TSpecification_Item> new_vdivde_items;
	std::vector<TDatabaseItem::TSpecification_Item> new_cmma_items;
	std::vector<TDatabaseItem::TSpecification_Item>::iterator item_iter;
	QString								expression;
	QString								b89_repeat_expression;
	double								b89_repeat_value(0.0);
	double								expression_values[5];
	int									expression_identification;
	int									standard_type_id;
	int									component_type_id;
	int									count;
	int									cntr;
	bool								valid;
	bool								b89_repeat_expression_valid(false);
		
	
	standard_type_id = this->Standard_Type_Id(standard_type);
	
	if(standard_type_id < 0)
	{
		return false;
	}

	database_item = this->Database_Item(make_id,model_id,option_type_id,standard_type_id,&valid);
	
	if(!valid)
	{
		return false;
	}
	
	count = database_item.Specification_Items_Count();

	for(cntr = 0;cntr < count;++cntr)
	{
		specification_item = database_item.Specification_Item(cntr);
		
		specification_item.source = TDatabaseItem::SOURCE_ESTIMATED;
		
		expression_values[0] = specification_item.expression_item.Expression_Value(0,TExpressionItem::UNITS_MM);
		expression_values[1] = specification_item.expression_item.Expression_Value(1,TExpressionItem::UNITS_MM);
		expression_values[2] = specification_item.expression_item.Expression_Value(2,TExpressionItem::UNITS_MM);
		expression_values[3] = specification_item.expression_item.Expression_Value(3,TExpressionItem::UNITS_MM);
		expression_values[4] = specification_item.expression_item.Expression_Value(4,TExpressionItem::UNITS_MM);
		
		if(standard_type.compare(QStringLiteral("B89.4.1")) == 0)
		{
			b89_repeat_expression = database_item.Specification_Expression(QStringLiteral("Repeat"),TExpressionItem::UNITS_MM);
			
			if(b89_repeat_expression.length())
			{
				b89_repeat_expression_valid = true;
				b89_repeat_value = b89_repeat_expression.toDouble();
			}
			
			if(specification_item.component.compare(QStringLiteral("Volume")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
								
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						
						if(b89_repeat_expression_valid)
						{
							expression = QString("%1+%2L/%3")
									.arg(b89_repeat_value,0,'f',4)
									.arg(expression_values[0] / 2.0,0,'f',4)
									.arg(expression_values[1]);

							specification_item.component = QStringLiteral("E0");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_10360_items.push_back(specification_item);
							
							specification_item.component = QStringLiteral("E150");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_10360_items.push_back(specification_item);
							
							specification_item.component = QStringLiteral("U3");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_vdivde_items.push_back(specification_item);
							
							specification_item.component = QStringLiteral("M");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_cmma_items.push_back(specification_item);
						}
						
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0] / 2.0,0,'f',4)
									.arg(expression_values[1] / 2.0,0,'f',4)
									.arg(expression_values[2]);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);

						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						
						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0] / 2.0,0,'f',4)
									.arg(expression_values[1] / 2.0,0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3],0,'f',4);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);

						break;
						
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						if(b89_repeat_expression_valid)
						{
							expression = QString("%1+(%2+%3Td)L/%4")
									.arg(b89_repeat_value,0,'f',4)
									.arg(expression_values[0] / 2.0,0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[3]);

							specification_item.component = QStringLiteral("E0");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_10360_items.push_back(specification_item);
							
							specification_item.component = QStringLiteral("E150");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_10360_items.push_back(specification_item);
							
							specification_item.component = QStringLiteral("U3");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_vdivde_items.push_back(specification_item);
							
							specification_item.component = QStringLiteral("M");
							specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
							new_cmma_items.push_back(specification_item);
						}
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:

						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0] / 2.0,0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2] / 2.0,0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("Repeat")) == 0)
			{

				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						
						specification_item.component = QStringLiteral("Rpt");
						expression = QString("%1").arg(expression_values[0],0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("Repeat");
						expression = QString("%1").arg(expression_values[0] * 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_cmma_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("R0");
						expression = QString("%1").arg(expression_values[0] * 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_10360_items.push_back(specification_item);

						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
		}
		else if(standard_type.compare(QStringLiteral("10360")) == 0)
		{
			if(specification_item.component.compare(QStringLiteral("E0")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2]);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);

						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3] * 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);

						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3],0,'f',4);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2] * 2.0,0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2],0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);

						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("E150")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2]);
						
						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3] * 2.0,0,'f',4);
							
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);

						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3],0,'f',4);
						
						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2] * 2.0,0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2],0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);
						
						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("R0")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						
						specification_item.component = QStringLiteral("Repeat");
						expression = QString("%1").arg(expression_values[0] / 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_b89_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("Repeat");
						expression = QString("%1").arg(expression_values[0],0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_cmma_items.push_back(specification_item);
						
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("Rpt")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:

						specification_item.component = QStringLiteral("Repeat");
						expression = QString("%1").arg(expression_values[0],0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_b89_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("Repeat");
						expression = QString("%1").arg(expression_values[0] * 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_cmma_items.push_back(specification_item);
						
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("Pftu")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:

						specification_item.component = QStringLiteral("V3");
						expression = QString("%1").arg(expression_values[0],0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_vdivde_items.push_back(specification_item);
						
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
		}
		else if(standard_type.compare(QStringLiteral("VDIVDE_2617")) == 0)
		{
			if(specification_item.component.compare(QStringLiteral("U3")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();

				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2]);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
					
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
					
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3] * 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);

						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3],0,'f',4);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2] * 2.0,0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2],0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("M");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_cmma_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("V3")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						
						specification_item.component = QStringLiteral("Pftu");
						expression = QString("%1").arg(expression_values[0],0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_10360_items.push_back(specification_item);

						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
		}
		else if(standard_type.compare(QStringLiteral("CMMA")) == 0)
		{
			if(specification_item.component.compare(QStringLiteral("M")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2L/%3")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2]);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1] * 2.0,0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3] * 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);

						expression = QString("%1+%2L/%3|%4")
									.arg(expression_values[0],0,'f',4)
									.arg(expression_values[1],0,'f',4)
									.arg(expression_values[2])
									.arg(expression_values[3],0,'f',4);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
						
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						
						specification_item.component = QStringLiteral("Volume");
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
									.arg(expression_values[0] * 2.0,0,'f',4)
									.arg(expression_values[1],0,'f',5)
									.arg(expression_values[2] * 2.0,0,'f',4)
									.arg(expression_values[3],0,'f',5)
									.arg(expression_values[4]);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_b89_items.push_back(specification_item);
						
						expression = QString("%1+%2Td+(%3+%4Td)L/%5")
										.arg(expression_values[0],0,'f',4)
										.arg(expression_values[1],0,'f',5)
										.arg(expression_values[2],0,'f',4)
										.arg(expression_values[3],0,'f',5)
										.arg(expression_values[4]);

						specification_item.component = QStringLiteral("E0");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("E150");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("U3");
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						new_vdivde_items.push_back(specification_item);
						
						break;
						
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
			}
			else if(specification_item.component.compare(QStringLiteral("Repeat")) == 0)
			{
				expression_identification = specification_item.expression_item.Expression_Type();
				
				switch(expression_identification)
				{
					case TExpressionItem::EXPRESSION_UNKNOWN:
						break;
					case TExpressionItem::EXPRESSION_VALUE:
						
						specification_item.component = QStringLiteral("Repeat");
						expression = QString("%1").arg(expression_values[0] / 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_b89_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("Rpt");
						expression = QString("%1").arg(expression_values[0] / 2.0,0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_10360_items.push_back(specification_item);
						
						specification_item.component = QStringLiteral("R0");
						expression = QString("%1").arg(expression_values[0],0,'f',4);
						
						specification_item.expression_item.Set_Expression(expression,TExpressionItem::UNITS_MM);
						
						new_10360_items.push_back(specification_item);
						
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
						break;
					case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
						break;
					case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
						break;
				}
 
			}
		}
	}
	
	QSqlQuery							query(*d_database);
	
	if(new_b89_items.size())
	{
		standard_type_id = this->Standard_Type_Id(QStringLiteral("B89.4.1"));
		
		for(item_iter = new_b89_items.begin();item_iter != new_b89_items.end();++item_iter)
		{
			component_type_id = this->Component_Type_Id(standard_type_id,(*item_iter).component);
			
			if(!this->Specification_Exists(make_id,model_id,option_type_id,standard_type_id,component_type_id))
			{
				if(!query.exec(QString("INSERT INTO specification (make_id,model_id,option_id,standard_id,component_id,expression,source) VALUES(%1,%2,%3,%4,%5,'%6',%7)")
							.arg(make_id)
							.arg(model_id)
							.arg(option_type_id)
							.arg(standard_type_id)
							.arg(component_type_id)
							.arg((*item_iter).expression_item.Expression(TExpressionItem::UNITS_MM))
							.arg((*item_iter).source)))
				{
					d_last_error = QString("Cannot add data to table 'specification'.  Reason: %1").arg(query.lastError().text());
					return false;
				}
			}
		}
	}
	
	if(new_10360_items.size())
	{
		standard_type_id = this->Standard_Type_Id(QStringLiteral("10360"));
		
		for(item_iter = new_10360_items.begin();item_iter != new_10360_items.end();++item_iter)
		{
			component_type_id = this->Component_Type_Id(standard_type_id,(*item_iter).component);
			
			if(!this->Specification_Exists(make_id,model_id,option_type_id,standard_type_id,component_type_id))
			{
				if(!query.exec(QString("INSERT INTO specification (make_id,model_id,option_id,standard_id,component_id,expression,source) VALUES(%1,%2,%3,%4,%5,'%6',%7)")
							.arg(make_id)
							.arg(model_id)
							.arg(option_type_id)
							.arg(standard_type_id)
							.arg(component_type_id)
							.arg((*item_iter).expression_item.Expression(TExpressionItem::UNITS_MM))
							.arg((*item_iter).source)))
				{
					d_last_error = QString("Cannot add data to table 'specification'.  Reason: %1").arg(query.lastError().text());
					return false;
				}
			}
		}
	}
	
	if(new_vdivde_items.size())
	{
		standard_type_id = this->Standard_Type_Id(QStringLiteral("VDIVDE_2617"));
		
		for(item_iter = new_vdivde_items.begin();item_iter != new_vdivde_items.end();++item_iter)
		{
			component_type_id = this->Component_Type_Id(standard_type_id,(*item_iter).component);
			
			if(!this->Specification_Exists(make_id,model_id,option_type_id,standard_type_id,component_type_id))
			{
				if(!query.exec(QString("INSERT INTO specification (make_id,model_id,option_id,standard_id,component_id,expression,source) VALUES(%1,%2,%3,%4,%5,'%6',%7)")
							.arg(make_id)
							.arg(model_id)
							.arg(option_type_id)
							.arg(standard_type_id)
							.arg(component_type_id)
							.arg((*item_iter).expression_item.Expression(TExpressionItem::UNITS_MM))
							.arg((*item_iter).source)))
				{
					d_last_error = QString("Cannot add data to table 'specification'.  Reason: %1").arg(query.lastError().text());
					return false;
				}
			}
		}
	}
	
	if(new_cmma_items.size())
	{
		standard_type_id = this->Standard_Type_Id(QStringLiteral("CMMA"));
		
		for(item_iter = new_cmma_items.begin();item_iter != new_cmma_items.end();++item_iter)
		{
			component_type_id = this->Component_Type_Id(standard_type_id,(*item_iter).component);
						
			if(!this->Specification_Exists(make_id,model_id,option_type_id,standard_type_id,component_type_id))
			{
				if(!query.exec(QString("INSERT INTO specification (make_id,model_id,option_id,standard_id,component_id,expression,source) VALUES(%1,%2,%3,%4,%5,'%6',%7)")
							.arg(make_id)
							.arg(model_id)
							.arg(option_type_id)
							.arg(standard_type_id)
							.arg(component_type_id)
							.arg((*item_iter).expression_item.Expression(TExpressionItem::UNITS_MM))
							.arg((*item_iter).source)))
				{
					d_last_error = QString("Cannot add data to table 'specification'.  Reason: %1").arg(query.lastError().text());
					return false;
				}
			}
		}
	}
	
	return true;
}

