/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <QStringList>
#include <assert.h>

#include "expressionitem.h"

TExpressionItem::TExpressionItem(void)
{
	d_expression_type = TExpressionItem::EXPRESSION_UNKNOWN;
	
	memset(d_expression_components,0,sizeof(d_expression_components));
}

TExpressionItem::TExpressionItem(
	const TExpressionItem					&rhs)
{
	if(&rhs != this)
	{
		d_expression_type = rhs.d_expression_type;
		memcpy(d_expression_components,rhs.d_expression_components,sizeof(d_expression_components));
	}
}

TExpressionItem& TExpressionItem::operator=(
	const TExpressionItem					&rhs)
{
	if(&rhs != this)
	{
		d_expression_type = rhs.d_expression_type;
		memcpy(d_expression_components,rhs.d_expression_components,sizeof(d_expression_components));
	}
	
	return *this;
}

TExpressionItem::~TExpressionItem(void)
{
}

QString TExpressionItem::Expression(
	const TExpressionItem::TUnits		units) const
{
	QString								result;
	
	if(units == TExpressionItem::UNITS_MM)
	{
		switch(d_expression_type)
		{
			case TExpressionItem::EXPRESSION_UNKNOWN:
				result = QStringLiteral("0.0000");
				break;
				
			case TExpressionItem::EXPRESSION_VALUE:
				result = QString("%1")
					.arg(d_expression_components[0],0,'f',4);
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_LENGTH:
				result = QString("%1/%2")
					.arg(d_expression_components[0],0,'f',4).arg(d_expression_components[1]);
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
				result = QString("%1+%2L/%3")
					.arg(d_expression_components[0],0,'f',4)
					.arg(d_expression_components[1],0,'f',4)
					.arg(d_expression_components[2]);
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
				result = QString("%1+%2L/%3|%4")
					.arg(d_expression_components[0],0,'f',4)
					.arg(d_expression_components[1],0,'f',4)
					.arg(d_expression_components[2])
					.arg(d_expression_components[3],0,'f',4);
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
				result = QString("(%1+%2Td)/%3")
					.arg(d_expression_components[0],0,'f',4)
					.arg(d_expression_components[1],0,'f',5)
					.arg(d_expression_components[2]);
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
				result = QString("%1+%2Td+(%3+%4Td)L/%5")
					.arg(d_expression_components[0],0,'f',4)
					.arg(d_expression_components[1],0,'f',5)
					.arg(d_expression_components[2],0,'f',4)
					.arg(d_expression_components[3],0,'f',5)
					.arg(d_expression_components[4]);
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
				result = QString("%1 C")
					.arg(d_expression_components[0],0,'f',2);
				break;
		}
	}
	else if(units == TExpressionItem::UNITS_UM)
	{
		switch(d_expression_type)
		{
			case TExpressionItem::EXPRESSION_UNKNOWN:
				result = QStringLiteral("0.0");
				break;
				
			case TExpressionItem::EXPRESSION_VALUE:
				result = QString("%1")
					.arg(d_expression_components[0] * 1000.0,0,'f',1);
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_LENGTH:
				result = QString("%1/%2")
					.arg(d_expression_components[0] * 1000.0,0,'f',1).arg(d_expression_components[1]);
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
				result = QString("%1+%2L/%3")
					.arg(d_expression_components[0] * 1000.0,0,'f',1)
					.arg(d_expression_components[1] * 1000.0,0,'f',1)
					.arg(d_expression_components[2]);
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
				result = QString("%1+%2L/%3|%4")
					.arg(d_expression_components[0] * 1000.0,0,'f',1)
					.arg(d_expression_components[1] * 1000.0,0,'f',1)
					.arg(d_expression_components[2])
					.arg(d_expression_components[3] * 1000.0,0,'f',1);
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
				result = QString("(%1+%2Td)/%3")
					.arg(d_expression_components[0] * 1000.0,0,'f',1)
					.arg(d_expression_components[1] * 1000.0,0,'f',2)
					.arg(d_expression_components[2]);
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
				result = QString("%1+%2Td+(%3+%4Td)L/%5")
					.arg(d_expression_components[0] * 1000.0,0,'f',1)
					.arg(d_expression_components[1] * 1000.0,0,'f',2)
					.arg(d_expression_components[2] * 1000.0,0,'f',1)
					.arg(d_expression_components[3] * 1000.0,0,'f',2)
					.arg(d_expression_components[4]);
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
				result = QString("%1 C")
					.arg(d_expression_components[0],0,'f',2);
				break;
		}
	}
	
	return result;
}

int TExpressionItem::Expression_Value_Count(void)
{
	switch(d_expression_type)
	{
		case TExpressionItem::EXPRESSION_UNKNOWN:
			return 0;
			
		case TExpressionItem::EXPRESSION_VALUE:
			return 1;
			
		case TExpressionItem::EXPRESSION_VALUE_LENGTH:
			return 2;
			
		case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
			return 3;
			
		case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
			return 4;
			
		case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
			return 3;
			
		case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
			return 5;
			
		case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
			return 1;
	}
	
	return 0;
}

double TExpressionItem::Expression_Value(
	const int							index,
	const int							units) const
{
	assert(static_cast<unsigned int>(index) < static_cast<unsigned int>(TExpressionItem::EXPRESSION_COMPONENT_SIZE));
	
	if(units == TExpressionItem::UNITS_UM)
	{
		if(units == TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS)
		{
			return d_expression_components[index];
		}
		
		return d_expression_components[index] * 1000.0;
	}

	return d_expression_components[index];
}

int TExpressionItem::Set_Expression(
	const QString						&expression,
	const TExpressionItem::TUnits		units)
{
	
	this->Evaluate_Expression(expression);
		
	if(units == TExpressionItem::UNITS_UM)
	{
		switch(d_expression_type)
		{
			case TExpressionItem::EXPRESSION_UNKNOWN:
				break;
				
			case TExpressionItem::EXPRESSION_VALUE:
				d_expression_components[0] /= 1000.0;
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_LENGTH:
				d_expression_components[0] /= 1000.0;
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH:
				d_expression_components[0] /= 1000.0;
				d_expression_components[1] /= 1000.0;
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT:
				d_expression_components[0] /= 1000.0;
				d_expression_components[1] /= 1000.0;
				d_expression_components[3] /= 1000.0;
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE:
				d_expression_components[0] /= 1000.0;
				d_expression_components[1] /= 1000.0;
				break;
				
			case TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE:
				d_expression_components[0] /= 1000.0;
				d_expression_components[1] /= 1000.0;
				d_expression_components[2] /= 1000.0;
				d_expression_components[3] /= 1000.0;
				break;
				
			case TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS:
				break;
		}
	}
	
	return d_expression_type;
}

void TExpressionItem::Evaluate_Expression(
	const QString						&expression)
{
	QRegularExpression					value_exp("^([0-9.]{1,9})$");
	QRegularExpression					temperature_celsius_exp("^([0-9.]{1,9})[C,c]$");
	QRegularExpression					temperature_fahrenheit_exp("^([0-9.]{1,9})[F,f]$");
	QRegularExpression					length_exp("^([0-9.]{1,9})[/]([0-9]{1,5})$");
	QRegularExpression					const_and_length_exp("^([0-9.]{1,9})[+]([0-9.]{1,9})[L][/]([0-9]{1,5})$");
	QRegularExpression					const_and_length_shortform_exp("^([0-9.]{1,9})[+][L][/]([0-9]{1,5})$");
	QRegularExpression					const_and_length_limit_exp("^([0-9.]{1,9})[+]([0-9.]{1,9})[L][/]([0-9]{1,5})[|]([0-9.]{1,9})$");
	QRegularExpression					const_and_length_limit_shortform_exp("^([0-9.]{1,9})[+][L][/]([0-9]{1,5})[|]([0-9.]{1,9})$");
	QRegularExpression					length_temperature_exp("^[(]([0-9.]{1,9})[+]([0-9.]{1,9})[T][d][)][/]([0-9]{1,5})$");
	QRegularExpression					const_and_length_temp_exp("^([0-9.]{1,9})[+]([0-9.]{1,9})[T][d][+][(]([0-9.]{1,9})[+]([0-9.]{1,9})[T][d][)][L][/]([0-9]{1,5})$");
	QRegularExpressionMatch				expression_match;
	QString								expression_text(expression);
	QStringList							expression_text_list;
	static const double					MINIMUM_TEMPERATURE(10.0);
	static const double					MAXIMUM_TEMPERATURE(40.0);
	
	d_expression_type = TExpressionItem::EXPRESSION_UNKNOWN;
	
	expression_text.remove(' ');		// remove any spaces
	expression_text.replace('l','L');	// just in case a lower case L is provided
	
	expression_match = const_and_length_temp_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 6)
		{			
			d_expression_type = TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_TEMPERATURE;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = expression_text_list[2].toDouble();
			d_expression_components[2] = expression_text_list[3].toDouble();
			d_expression_components[3] = expression_text_list[4].toDouble();
			d_expression_components[4] = expression_text_list[5].toDouble();
			
			return;
		}
	}
	
	expression_match = const_and_length_limit_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 5)
		{
			d_expression_type = TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = expression_text_list[2].toDouble();
			d_expression_components[2] = expression_text_list[3].toDouble();
			d_expression_components[3] = expression_text_list[4].toDouble();

			return;
		}
	}
	
	expression_match = const_and_length_limit_shortform_exp.match(expression_text);		// format is 3+L/333|3 with implying 3+1L/333|3
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 4)
		{
			d_expression_type = TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH_LIMIT;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = 1.0;
			d_expression_components[2] = expression_text_list[2].toDouble();
			d_expression_components[3] = expression_text_list[3].toDouble();
			
			return;
		}
	}
	
	expression_match = length_temperature_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 4)
		{
			d_expression_type = TExpressionItem::EXPRESSION_VALUE_LENGTH_TEMPERATURE;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = expression_text_list[2].toDouble();
			d_expression_components[2] = expression_text_list[3].toDouble();
			
			return;
		}
	}
	
	expression_match = const_and_length_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 4)
		{
			d_expression_type = TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = expression_text_list[2].toDouble();
			d_expression_components[2] = expression_text_list[3].toDouble();
			
			return;
		}
	}
	
	expression_match = const_and_length_shortform_exp.match(expression_text);	// format is 3+L/333 with implying 3+1L/333
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 3)
		{
			d_expression_type = TExpressionItem::EXPRESSION_CONST_VALUE_LENGTH;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = 1.0;
			d_expression_components[2] = expression_text_list[2].toDouble();
			
			return;
		}
	}
	
	expression_match = length_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 3)
		{
			d_expression_type = TExpressionItem::EXPRESSION_VALUE_LENGTH;
			
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_components[1] = expression_text_list[2].toDouble();
			
			return;
		}
	}
	
	expression_match = temperature_celsius_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 2)
		{
			d_expression_components[0] = expression_text_list[1].toDouble();
			
			if(d_expression_components[0] > MINIMUM_TEMPERATURE && d_expression_components[0] < MAXIMUM_TEMPERATURE)
			{
				d_expression_type = TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS;
			}
		}
		
		return;
	}
	
	expression_match = temperature_fahrenheit_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 2)
		{
			d_expression_components[0] = (expression_text_list[1].toDouble() - 32.0) * 5.0 / 9.0;	// temperature provided with 'F', convert to celsius

			if(d_expression_components[0] > MINIMUM_TEMPERATURE && d_expression_components[0] < MAXIMUM_TEMPERATURE)
			{
				d_expression_type = TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS;
			}
		}
		
		return;
	}
	
	expression_match = value_exp.match(expression_text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 2)
		{
			d_expression_components[0] = expression_text_list[1].toDouble();
			d_expression_type = TExpressionItem::EXPRESSION_VALUE;
		}
		
		return;
	}
}
