/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QFile>
#include <QTextStream>
#include <QByteArray>

#include "database.h"
#include "import_export.h"


TImportExportThread::TImportExportThread(
	const QObject						*parent)
:QThread(const_cast<QObject*>(parent))
{
	d_database = 0;
	d_execution_mode = TImportExportThread::MODE_EXPORT;
}

TImportExportThread::~TImportExportThread(void)
{
}

void TImportExportThread::run(void)
{
	if(!d_database)
	{
		return;
	}
	
	d_error_state = false;
	
	switch(d_execution_mode)
	{
		case TImportExportThread::MODE_EXPORT:
			this->Export_Database();
			break;
			
		case TImportExportThread::MODE_IMPORT:
			this->Import_Database();
			break;
	}
}

void TImportExportThread::Export_Database(void)
{
	QFile								file;
	QString								entry_line;
	QString								text;
	QStringList							make_list;
	QStringList							model_list;
	QStringList							option_list;
	QStringList							standard_type_list;
	QStringList::const_iterator			make_iter;
	QStringList::const_iterator			model_iter;
	QStringList::const_iterator			option_iter;
	QStringList::const_iterator			standard_type_iter;
	TDatabaseItem						database_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	int									count;
	int									cntr;
	int									progress_range;
	int									progress_value;
	bool								valid;
	
	file.setFileName(d_file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		file.write("CMMSpecificationsDatabase:Version=1:Format=CSV\n\n");
		
		standard_type_list = d_database->Standard_Type_List(&valid);
		
		if(!valid)
		{
			file.close();
			
			d_last_error = d_database->Last_Error();
			d_error_state = true;

			return;
		}
		
		if(d_export_make_list.size())
		{
			make_list = d_export_make_list;
		}
		else
		{
			make_list = d_database->CMM_Make_List(&valid);
		}
		
		make_list.sort();
		
		progress_range = make_list.size();
		progress_value = 0;

		emit Update_Progress_Text(QStringLiteral("Exporting from database.  Please wait..."));

		if(valid)
		{
			for(make_iter = make_list.begin();make_iter != make_list.end();++make_iter)
			{
				model_list = d_database->CMM_Model_List((*make_iter),&valid);
				progress_value++;
				
				if(valid)
				{
					for(model_iter = model_list.begin();model_iter != model_list.end();++model_iter)
					{
						option_list = d_database->CMM_Option_List((*make_iter),(*model_iter),&valid);
						
						if(valid)
						{
							for(option_iter = option_list.begin();option_iter != option_list.end();++option_iter)
							{
								for(standard_type_iter = standard_type_list.begin();standard_type_iter != standard_type_list.end();++standard_type_iter)
								{
									database_item = d_database->Database_Item((*make_iter),(*model_iter),(*option_iter),(*standard_type_iter),&valid);
									
									if(valid)
									{
										database_item.Clear_Estimated_Specification_Items();
										
										count = database_item.Specification_Items_Count();
										
										if(count)
										{
											entry_line = QString("%1,%2,%3,%4").arg(*make_iter).arg(*model_iter).arg(*option_iter).arg(*standard_type_iter);
											
											for(cntr = 0;cntr < count;++cntr)
											{
												specification_item = database_item.Specification_Item(cntr);
												
												if(specification_item.source == TDatabaseItem::SOURCE_MANUFACTURER)
												{
													entry_line += QString(",%1:%2")
														.arg(specification_item.component)
														.arg(specification_item.expression_item.Expression(TExpressionItem::UNITS_MM));
												}
											}
											
											entry_line += QStringLiteral("\n");
											file.write(entry_line.toLatin1());
										}
									}
								}
							}
						}
					}
				}
				
				emit Update_Progress(progress_value,progress_range);
			}
		}
		file.close();
	}
}

void TImportExportThread::Import_Database(void)
{
	QByteArray							file_data;
	QFile								file;
	QString								line;
	QString								text;
	QString								make;
	QString								model;
	QString								option;
	QString								standard_type;
	QStringList							line_items;
	QStringList							spec_items;
	QStringList							make_list;
	QStringList							model_list;
	QStringList							option_list;
	QStringList::const_iterator			make_iter;
	QStringList::const_iterator			model_iter;
	QStringList::const_iterator			option_iter;
	TDatabaseItem						database_item;
	TDatabaseItem::TSpecification_Item	specification_item;
	int									count;
	int									cntr;
	int									progress_range;
	int									progress_value(0);
	bool								init(true);
	bool								valid;
	
	file.setFileName(d_file_name);
	
	if(file.open(QIODevice::ReadOnly | QIODevice::Text))
	{
		file_data = file.readAll();
		file.close();
		
		progress_range = file_data.count('\n');
		progress_value = 0;
		
		emit Update_Progress_Text(QStringLiteral("Importing to database.  Please wait..."));
		
		QTextStream						stream(file_data);
		
		while(!stream.atEnd())
		{
			line = stream.readLine().simplified();
			progress_value++;
			
			if(line.size() > 0 && line.startsWith('#') == false)
			{
				if(init)
				{
					if(!line.startsWith(QStringLiteral("CMMSpecificationsDatabase:Version=1:Format=CSV"),Qt::CaseInsensitive))
					{
						d_last_error = QStringLiteral("Imported CSV file does not start with the proper header line.");
						d_error_state = true;
						
						return;
					}
					
					init = false;
				}
				else
				{
					line_items = line.split(',');
					count = line_items.size();
					
					if(count > 4)	// must be at least one entry past make,model,option,standard
					{
						make = line_items[0];
						model = line_items[1];
						option = line_items[2];
						standard_type = line_items[3];
						
						database_item.Set_Make(make);
						database_item.Set_Model(model);
						database_item.Set_Option(option);
						database_item.Set_Standard_Type(standard_type);
						
						database_item.Clean_Strings();
						
						specification_item.source = TDatabaseItem::SOURCE_MANUFACTURER;
						
						database_item.Clear_Specification_Items();
						
						for(cntr = 4;cntr < count;++cntr)
						{
							spec_items = line_items[cntr].split(':');
							
							if(spec_items.size() == 2)
							{
								specification_item.component = spec_items[0];
								
								specification_item.expression_item.Set_Expression(spec_items[1],TExpressionItem::UNITS_MM);
								database_item.Add_Specification_Item(specification_item);
							}
						}
						
						d_database->Add_Database_Item(database_item);
						
						if(!make_list.contains(make))
						{
							make_list.push_back(make);
						}
					}
				}
				
				emit Update_Progress(progress_value,progress_range);
			}
		}
	}
	
//	make_list = d_database->CMM_Make_List(&valid);
	
	progress_range = make_list.size() + 1;
	progress_value = 0;
	
	emit Update_Progress_Text(QStringLiteral("Estimating Specifications.  Please wait..."));

	for(make_iter = make_list.begin();make_iter != make_list.end();++make_iter)
	{
		progress_value++;
		model_list = d_database->CMM_Model_List((*make_iter),&valid);
		
		for(model_iter = model_list.begin();model_iter != model_list.end();++model_iter)
		{

			option_list = d_database->CMM_Option_List((*make_iter),(*model_iter),&valid);

			for(option_iter = option_list.begin();option_iter != option_list.end();++option_iter)
			{

				d_database->Create_Estimated_Specifications((*make_iter),(*model_iter),(*option_iter));
			}
		}
		
		emit Update_Progress(progress_value,progress_range);
	}
}

