/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QSpacerItem>
#include <QTableWidget>
#include <QTableWidgetItem>
#include <QSizePolicy>
#include <assert.h>

#include "database.h"
#include "databaseitem.h"
#include "expressionitem.h"

#include "specificationeditordialog.h"

static const int						TABLE_COLUMN_COUNT = 4;

TSpecificationEditorDialog::TSpecificationEditorDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*lower_separator_line;
	QFrame								*upper_separator_line;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*button_hlayout;
	QLabel								*make_label;
	QLabel								*model_label;
	QLabel								*option_label;
	QPushButton							*cancel_button;
	QSpacerItem							*button_hspacer;
	QSizePolicy							sizepolicy_maximum_preferred(QSizePolicy::Maximum, QSizePolicy::Preferred);
	QSizePolicy							sizepolicy_maximumexpanding_preferred(QSizePolicy::MinimumExpanding, QSizePolicy::Preferred);
	
	sizepolicy_maximum_preferred.setHorizontalStretch(0);
	sizepolicy_maximum_preferred.setVerticalStretch(0);
	
	sizepolicy_maximumexpanding_preferred.setHorizontalStretch(0);
	sizepolicy_maximumexpanding_preferred.setVerticalStretch(0);

	this->resize(601,801);

	dialog_layout = new QGridLayout(this);

	make_label = new QLabel(this);
	sizepolicy_maximum_preferred.setHeightForWidth(make_label->sizePolicy().hasHeightForWidth());
	make_label->setSizePolicy(sizepolicy_maximum_preferred);
	dialog_layout->addWidget(make_label,0,0,1,1);
	
	d_make_edit = new QLineEdit(this);
	d_make_edit->setAlignment(Qt::AlignCenter);
	sizepolicy_maximumexpanding_preferred.setHeightForWidth(d_make_edit->sizePolicy().hasHeightForWidth());
	d_make_edit->setSizePolicy(sizepolicy_maximumexpanding_preferred);
	dialog_layout->addWidget(d_make_edit,0,1,1,1);
	
	d_make_exists_label = new QLabel(this);
	sizepolicy_maximum_preferred.setHeightForWidth(d_make_exists_label->sizePolicy().hasHeightForWidth());
	d_make_exists_label->setSizePolicy(sizepolicy_maximum_preferred);
	d_make_exists_label->setMinimumSize(QSize(80, 0));
	dialog_layout->addWidget(d_make_exists_label,0,2,1,1);

	model_label = new QLabel(this);
	sizepolicy_maximum_preferred.setHeightForWidth(model_label->sizePolicy().hasHeightForWidth());
	model_label->setSizePolicy(sizepolicy_maximum_preferred);
	dialog_layout->addWidget(model_label,1,0,1,1);

	d_model_edit = new QLineEdit(this);
	d_model_edit->setAlignment(Qt::AlignCenter);
	sizepolicy_maximumexpanding_preferred.setHeightForWidth(d_model_edit->sizePolicy().hasHeightForWidth());
	d_model_edit->setSizePolicy(sizepolicy_maximumexpanding_preferred);
	dialog_layout->addWidget(d_model_edit,1,1,1,1);
	
	d_model_exists_label = new QLabel(this);
	sizepolicy_maximum_preferred.setHeightForWidth(d_model_exists_label->sizePolicy().hasHeightForWidth());
	d_model_exists_label->setSizePolicy(sizepolicy_maximum_preferred);
	d_model_exists_label->setMinimumSize(QSize(80, 0));
	dialog_layout->addWidget(d_model_exists_label,1,2,1,1);

	option_label = new QLabel(this);
	sizepolicy_maximum_preferred.setHeightForWidth(option_label->sizePolicy().hasHeightForWidth());
	option_label->setSizePolicy(sizepolicy_maximum_preferred);
	dialog_layout->addWidget(option_label,2,0,1,1);

	d_option_edit = new QLineEdit(this);
	d_option_edit->setAlignment(Qt::AlignCenter);
	sizepolicy_maximumexpanding_preferred.setHeightForWidth(d_option_edit->sizePolicy().hasHeightForWidth());
	d_option_edit->setSizePolicy(sizepolicy_maximumexpanding_preferred);
	dialog_layout->addWidget(d_option_edit,2,1,1,1);
	
	d_option_exists_label = new QLabel(this);
	sizepolicy_maximum_preferred.setHeightForWidth(d_option_exists_label->sizePolicy().hasHeightForWidth());
	d_option_exists_label->setSizePolicy(sizepolicy_maximum_preferred);
	d_option_exists_label->setMinimumSize(QSize(80, 0));
	dialog_layout->addWidget(d_option_exists_label,2,2,1,1);
	
	upper_separator_line = new QFrame(this);
	upper_separator_line->setFrameShape(QFrame::HLine);
	upper_separator_line->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(upper_separator_line,3,0,1,3);

	d_specification_table = new QTableWidget(this);
	dialog_layout->addWidget(d_specification_table,4,0,1,3);

	lower_separator_line = new QFrame(this);
	lower_separator_line->setFrameShape(QFrame::HLine);
	lower_separator_line->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(lower_separator_line,5,0,1,3);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	cancel_button = new QPushButton(this);
	cancel_button->setAutoDefault(false);
	button_hlayout->addWidget(cancel_button);

	d_accept_button = new QPushButton(this);
	d_accept_button->setAutoDefault(false);
	button_hlayout->addWidget(d_accept_button);
	dialog_layout->addLayout(button_hlayout,6,0,1,3);
	
	// defaults
	
	d_database = 0;

	this->setWindowTitle(QStringLiteral("Specification Editor"));
	
	make_label->setText(QStringLiteral("Make:"));
	model_label->setText(QStringLiteral("Model:"));
	option_label->setText(QStringLiteral("Option:"));
	cancel_button->setText(QStringLiteral("Cancel"));
	d_accept_button->setText(QStringLiteral("Accept"));
	
	connect(d_accept_button,&QPushButton::clicked,this,&TSpecificationEditorDialog::accept);
	connect(cancel_button,&QPushButton::clicked,this,&TSpecificationEditorDialog::reject);
	
	connect(d_specification_table,&QTableWidget::cellChanged,this,&TSpecificationEditorDialog::Cell_Changed);
	connect(d_make_edit,&QLineEdit::editingFinished,this,&TSpecificationEditorDialog::Identification_Changed);
	connect(d_model_edit,&QLineEdit::editingFinished,this,&TSpecificationEditorDialog::Identification_Changed);
	connect(d_option_edit,&QLineEdit::editingFinished,this,&TSpecificationEditorDialog::Identification_Changed);
}

TSpecificationEditorDialog::~TSpecificationEditorDialog(void)
{
}

QString TSpecificationEditorDialog::Make(void) const
{
	return d_make_edit->text().trimmed();
}

QString TSpecificationEditorDialog::Model(void) const
{
	return d_model_edit->text().trimmed();
}

QString TSpecificationEditorDialog::Option(void) const
{
	return d_option_edit->text().trimmed();
}

std::vector<TSpecificationEditorDialog::TEntryItem> TSpecificationEditorDialog::Specification_Entries(void)
{
	std::vector<TSpecificationEditorDialog::TEntryItem> specification_items;
	std::vector<TSpecificationEditorDialog::TEntryItem>::const_iterator iter;
	
	for(iter = d_entry_items.begin();iter != d_entry_items.end();++iter)
	{
		if((*iter).valid && (*iter).expression.length())
		{
			specification_items.push_back(*iter);
		}
	}
	
	return specification_items;
}

void TSpecificationEditorDialog::Reset(
	TDatabase							* const database,
	const QString						&make,
	const QString						&model,
	const QString						&option,
	const std::vector<TEntryItem>		&entry_items)
{
	QTableWidgetItem					*table_item;
	TEntryItem							entry_item;
	int									count;
	int									cntr;
	bool								prev_state;
	
	assert(TABLE_COLUMN_COUNT > 3);
	assert(database);
	
	d_database = database;
	
	prev_state = d_specification_table->blockSignals(true);
	
	d_entry_items = entry_items;

	d_make_edit->setText(make);
	d_model_edit->setText(model);
	d_option_edit->setText(option);
	
	make.length() ? d_make_exists_label->setText(QStringLiteral("Exists")) : d_make_exists_label->setText(QString());
	model.length() ? d_model_exists_label->setText(QStringLiteral("Exists")) : d_model_exists_label->setText(QString());
	option.length() ? d_option_exists_label->setText(QStringLiteral("Exists")) : d_option_exists_label->setText(QString());
	
	d_specification_table->clear();
	this->Create_Table_Columns();
	
	count = static_cast<int>(entry_items.size());
	d_specification_table->setRowCount(count);
	
	for(cntr = 0;cntr < count;++cntr)
	{
		entry_item = d_entry_items[cntr];
		
		d_entry_items[cntr].valid = 1;	// assume all provided expressions are valid

		table_item = new QTableWidgetItem();
		table_item->setText(entry_item.standard);
		table_item->setFlags(table_item->flags()^(Qt::ItemIsSelectable|Qt::ItemIsEditable));
		d_specification_table->setItem(cntr,0,table_item);
		
		table_item = new QTableWidgetItem();
		table_item->setText(entry_item.component);
		table_item->setFlags(table_item->flags()^(Qt::ItemIsSelectable|Qt::ItemIsEditable));
		d_specification_table->setItem(cntr,1,table_item);
		
		table_item = new QTableWidgetItem();
		table_item->setText(entry_item.expression);
		d_specification_table->setItem(cntr,2,table_item);
		
		table_item = new QTableWidgetItem();	// status is blank by default
		table_item->setFlags(table_item->flags()^(Qt::ItemIsSelectable|Qt::ItemIsEditable));
		d_specification_table->setItem(cntr,3,table_item);
	}
	
	this->Update_Status_Column();
	
	d_specification_table->blockSignals(prev_state);
	
	d_specification_table->resizeColumnsToContents();
	
	d_accept_button->setEnabled(false);
}

void TSpecificationEditorDialog::Cell_Changed(
	int									row,
	int									column)
{
	Q_UNUSED(column);
	
	this->Update_Row_Status(row);
	this->Update_Accept_Button_Enable();
}

void TSpecificationEditorDialog::Identification_Changed(void)
{
	this->Update_Accept_Button_Enable();
	this->Update_Exist_Labels();
}

void TSpecificationEditorDialog::Create_Table_Columns(void)
{
	QTableWidgetItem					*table_column_item;
	int									cntr;
	
	d_specification_table->setColumnCount(TABLE_COLUMN_COUNT);
	
	for(cntr = 0;cntr < TABLE_COLUMN_COUNT;++cntr)
	{
		table_column_item = new QTableWidgetItem();
		d_specification_table->setHorizontalHeaderItem(cntr,table_column_item);
		
		switch(cntr)
		{
			case 0:
				table_column_item->setText(QStringLiteral("Standard"));
				break;
				
			case 1:
				table_column_item->setText(QStringLiteral("Component"));
				break;
				
			case 2:
				table_column_item->setText(QStringLiteral("Specification Expression"));
				break;
				
			case 3:
				table_column_item->setText(QStringLiteral("Expression Status"));
				break;
		}
	}
}

void TSpecificationEditorDialog::Update_Status_Column(void)
{
	QTableWidgetItem					*table_item;
	TEntryItem							entry_item;
	QString								expression;
	int									expression_type;
	int									count;
	int									cntr;
	TExpressionItem						expression_item;
	
	count = d_specification_table->rowCount();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		table_item = d_specification_table->item(cntr,2);	// expression column
		
		expression = table_item->text().trimmed();
		
		table_item = d_specification_table->item(cntr,3);	// status column

		if(expression.length())
		{
			entry_item = d_entry_items[cntr];
			
			expression_type = expression_item.Set_Expression(expression,TExpressionItem::UNITS_UM);
			
			if(expression_type & entry_item.expression_types)
			{
				table_item->setText(QStringLiteral("Ok"));
				d_entry_items[cntr].valid = 1;
			}
			else
			{
				table_item->setText(QStringLiteral("Invalid"));
				d_entry_items[cntr].valid = 0;
			}
		}
		else
		{
			table_item->setText(QString());
		}
	}
}

void TSpecificationEditorDialog::Update_Row_Status(
	const int							row)
{
	QTableWidgetItem					*table_item;
	TEntryItem							entry_item;
	QString								expression;
	int									expression_type;
	TExpressionItem						expression_item;
	
	if(row < d_specification_table->rowCount() &&
	   row < static_cast<int>(d_entry_items.size()))
	{
		table_item = d_specification_table->item(row,2);	// expression column
		
		expression = table_item->text().trimmed();
		
		table_item = d_specification_table->item(row,3);	// status column
		
		if(expression.length())
		{
			entry_item = d_entry_items[row];
			
			if(entry_item.expression_types & TExpressionItem::EXPRESSION_VALUE_TEMPERATURE_CELSIUS)	// expression must end in 'C' to process
			{
				if(!(expression.endsWith('C') || expression.endsWith('c')))
				{
					expression.push_back('C');
				}
			}
			
			expression_type = expression_item.Set_Expression(expression,TExpressionItem::UNITS_UM);
			
			if(expression_type & entry_item.expression_types)
			{
				table_item->setText(QStringLiteral("Ok"));
				d_entry_items[row].valid = 1;
				
				d_entry_items[row].expression = expression;
			}
			else
			{
				table_item->setText(QStringLiteral("Invalid"));
				d_entry_items[row].valid = 0;
			}
		}
		else
		{
			table_item->setText(QString());
			d_entry_items[row].expression = QString();
		}
	}
}

void TSpecificationEditorDialog::Update_Accept_Button_Enable(void)
{
	QString								make;
	QString								model;
	QString								option;
	std::vector<TEntryItem>::const_iterator iter;
	
	make = d_make_edit->text().trimmed();
	model = d_model_edit->text().trimmed();
	option = d_option_edit->text().trimmed();
	
	d_accept_button->setEnabled(false);

	
	if(make.length() && model.length() && option.length())
	{
		for(iter = d_entry_items.begin();iter != d_entry_items.end();++iter)
		{
			if(!(*iter).valid)
			{
				return;
			}
		}
		
		d_accept_button->setEnabled(true);
	}

}

void TSpecificationEditorDialog::Update_Exist_Labels(void)
{
	QString								make;
	QString								model;
	QString								option;
	QString								clean_string;
	bool								make_exists(false);
	bool								model_exists(false);
	bool								option_exists(false);
	bool								prev_state;
	
	assert(d_database);
	
	make = d_make_edit->text().trimmed();
	
	if(make.length())
	{
		clean_string = TDatabaseItem::Clean_String(make);
		if(clean_string.compare(make) != 0)
		{
			make = clean_string;
			
			prev_state = d_make_edit->blockSignals(true);
			d_make_edit->setText(make);
			d_make_edit->blockSignals(prev_state);
		}

		make_exists = d_database->CMM_Make_Exists(make);
		
		model = d_model_edit->text().trimmed();

		if(model.length())
		{
			clean_string = TDatabaseItem::Clean_String(model);
			if(clean_string.compare(model) != 0)
			{
				model = clean_string;
				
				prev_state = d_model_edit->blockSignals(true);
				d_model_edit->setText(model);
				d_model_edit->blockSignals(prev_state);
			}

			if(make_exists)
			{
				model_exists = d_database->CMM_Model_Exists(make,model);
				
				option = d_option_edit->text().trimmed();
				
				if(option.length())
				{
					clean_string = TDatabaseItem::Clean_String(option);
					if(clean_string.compare(option) != 0)
					{
						option = clean_string;
						
						prev_state = d_option_edit->blockSignals(true);
						d_option_edit->setText(option);
						d_option_edit->blockSignals(prev_state);
					}

					if(make_exists && model_exists)
					{
						option_exists = d_database->CMM_Option_Exists(make,model,option);
					}
				}
			}
		}
	}
	
	make_exists ? d_make_exists_label->setText(QStringLiteral("Exists")) : d_make_exists_label->setText(QString());
	model_exists ? d_model_exists_label->setText(QStringLiteral("Exists")) : d_model_exists_label->setText(QString());
	option_exists ? d_option_exists_label->setText(QStringLiteral("Exists")) : d_option_exists_label->setText(QString());
}

